"""
Support for working with EPREM source files.
"""

import abc
import collections.abc
import json
import numbers
import pathlib
import re
import typing

from .. import container
from .. import paths
from .. import symbolic
from . import _metadata


class SourceFile(collections.abc.Mapping):
    """An object representing parameters in an EPREM source file."""

    _db_path = pathlib.Path(__file__).parent / 'reference.json'

    def __init__(
        self,
        db_key: str,
        filename: str,
        srcdir: typing.Union[str, pathlib.Path]=None,
    ) -> None:
        """Initialize an instance of this class."""
        path = self.build_path(filename, srcdir)
        if path and path.exists():
            loaded = self.load_from_source(paths.TextFile(path))
            self._origin = path
        else:
            loaded = self.load_from_package(db_key)
            self._origin = self._db_path
        self._definitions = self.standardize(loaded)

    @property
    def origin(self):
        """The path to definitions.
        
        This property represents the user-provided path to the source code used
        to build the instance definitions, or the path to the database of
        default definitions.
        """
        return self._origin

    def __iter__(self) -> typing.Iterator:
        """Called for iter(self)."""
        return iter(self.definitions)

    def __len__(self) -> int:
        """Called for len(self)."""
        return len(self.definitions)

    @property
    def definitions(self):
        """The parameter definitions in this file."""
        return self._definitions

    @abc.abstractmethod
    def standardize(self, loaded: dict):
        """Convert loaded attributes into standard definitions.
        
        This abstract method provides a default implementation that simply
        returns the input object.
        """
        return loaded

    def build_path(
        self,
        filename: str,
        source: paths.PathLike=None,
    ) -> typing.Optional[pathlib.Path]:
        """Create the full path to the source file, if possible."""
        try:
            path = paths.fullpath(source, strict=True)
            if path.is_dir():
                path /= filename
        except TypeError:
            path = None
        return path

    def load_from_package(self, key: str) -> dict:
        """Load argument definitions from the package defaults."""
        with self._db_path.open('r') as fp:
            loaded = json.load(fp)
        return loaded[key]

    @abc.abstractmethod
    def load_from_source(self, file: paths.TextFile) -> dict:
        """Load argument definitions from the source file."""
        pass

    def get(self, key: str, default: typing.Any=None, format: str=None):
        """Get (and optionally format) a parameter value or `default`.
        
        The default implementation is identical to `collections.abc.Mapping.get`
        when `format=None`, and otherwise raises an exception. Concrete
        subclasses may wish to define additional `format`-based behavior by
        overloading this method.
        """
        if not format:
            return super().get(key, default)
        raise NotImplementedError(f"Unrecognized format {format}")

    def todict(self, mode: str=None, sort: bool=False):
        """Convert this object to a `dict` according to `mode`.
        
        Parameters
        ----------
        mode : string, optional
            A defined formatting mode. This method will pass `mode` to the
            `format` keyword of `~get` for each value.

        sort : boolean, default=false
            If true, sort the parameters before creating the `dict`. The default
            behavior is to respect insertion order.
        """
        keys = sorted(self.keys()) if sort else self.keys()
        return {key: self.get(key, format=mode) for key in keys}

    def tolines(self, indent: int=0, **kwargs):
        """Convert this object to a `list` of `str`.

        Parameters
        ----------
        indent : int, default=0
            The number of spaces to indent each line.

        **kwargs
            Additional options to pass to `~todict`.
        """
        lines = []
        pad = indent * ' '
        selfdict = self.todict(**kwargs)
        for name, current in selfdict.items():
            if isinstance(current, typing.Mapping):
                value = '\n'.join(
                    f"{pad}{pad}{k!r}: {v if k == 'type' else repr(v)},"
                    for k, v in current.items()
                )
            else:
                value = current
            line = f"{pad}{name!r}" + ": {\n" + f"{value}\n{pad}" + "}"
            lines.append(line)
        return lines

    def tostring(self, **kwargs):
        """Convert this object to a formatted string.
        
        This method will pass all arguments to `~tolines`. The default
        implementation simply joins the result of `~tolines` via `str.join`.
        """
        lines = self.tolines(**kwargs)
        return '\n'.join(f"{line}," for line in lines)


class BaseTypeDef:
    """Pattern parser for defined constants in baseTypes.h."""

    @property
    def pattern(self):
        return re.compile(r"""
            # the start of the string (no following whitespace)
            \A
            # the `#define` pre-processor directive
            \#define
            # at least one whitespace character
            \s+
            # the defined name
            (?P<name>[0-9A-Z_]+)
            # at least one whitespace character
            \s+
            # the defined value
            (?P<value>[-+\.\*\/\d\w\s\(\)]+)
            # the end of the string (no trailing whitespace)
            \Z
        """, re.VERBOSE)

    def match(self, line: str):
        """Identify lines that define a runtime constant."""
        if match := self.pattern.match(line.strip()):
            return match.groupdict()

    def parse(self, parsable: typing.Dict[str, str]):
        """Parse a line that defines a runtime constant."""
        return parsable['name'], parsable['value']


class BaseTypesH(SourceFile):
    """A representation of constant values in EPREM `baseTypes.h`."""

    def __init__(self, srcdir: typing.Union[str, pathlib.Path]=None) -> None:
        super().__init__('BASETYPES_H', 'baseTypes.h', srcdir)
        self._types = None
        self._cache = {}

    def standardize(self, loaded: dict):
        return super().standardize(loaded)

    def load_from_source(self, file: paths.TextFile) -> dict:
        typedef = BaseTypeDef()
        return file.extract(typedef.match, typedef.parse)

    def get(self, key: str, default: typing.Any=None, format: str=None):
        if format == 'raw':
            value = self.definitions.get(key, default)
            if isinstance(value, symbolic.Expression):
                return value.format(separator=' * ')
            return value
        if format == 'full':
            value = self.get(key, default)
            result = {'value': value}
            for k, v in _metadata._BASETYPES_H.get(key, {}).items():
                result[k] = v.__name__ if k == 'type' else v
            return result
        return super().get(key, default, format)

    def tostring(self, **kwargs):
        s = super().tostring(**kwargs)
        return 'BASETYPES_H = {\n' + f"{s}" + '\n}'

    def __getitem__(self, key: str) -> numbers.Real:
        """Access constants by keyword."""
        if key in self._cache:
            return self._cache[key]
        if key in self.definitions:
            value = self._compute(key)
            self._cache[key] = value
            return value
        raise KeyError(f"No {key!r} in {self!r}")

    def _compute(self, key: str) -> numbers.Real:
        """Compute the value of a defined constant."""
        target = self.definitions[key]
        realtype = self.types.get(key)
        if isinstance(target, realtype):
            return target
        if any(c in target for c in {'*', '/', 'sqrt'}):
            return self._evaluate(symbolic.expression(target))
        if realtype:
            return realtype(target)
        raise TypeError(target)

    @property
    def types(self):
        """The type of each constant."""
        if self._types is None:
            self._types = {
                k: v['type'] for k, v in _metadata._BASETYPES_H.items()
            }
        return self._types

    def _evaluate(self, expression: symbolic.Expression):
        """Internal method for evaluating symbolic definitions."""
        value = 1.0
        for term in expression:
            if term.base in self.definitions:
                value *= float(term(self._compute(term.base)))
            elif term.base == '1':
                value *= float(term)
        return value


class FunctionCall:
    """Pattern parser for function calls."""

    # TODO: 
    # - __init__ should take arguments that allow the user to define possible
    #   function names (e.g., 'readInt', 'readDouble') and target-variable
    #   prefix (e.g., 'config').
    # - It may be better to just grab everything in parentheses as `args` and
    #   let `parse` sort them out.

    @property
    def pattern(self):
        return re.compile(r"""
            # the start of the string, followed by optional whitespace
            \A\s*
            # the assignment target (possibly a struct member)
            (?:config\.)?\w+
            # equals sign surrounded by optional whitespace
            \s*=\s*
            # kludge for `(char*)readString(...)`
            (?:\(char\*\))?
            # the name of the file-reading method
            (?P<mode>read(?:Bool|Int|Double|DoubleArray|String))
            # beginning of function call
            \(
                # the parameter key in the config file
                (?P<key>\"\w*\")
                # optional whitespace followed by a comma
                \s*\,
                # the remaining arguments
                (?P<args>.*?)
            # end of function call
            \)
            # C statement terminator
            \;
            # optional whitespace, followed by the end of the string
            \s*\Z
        """, re.VERBOSE)

    def match(self, line: str):
        """Identify lines that read config-file input."""
        if match := self.pattern.match(line.strip()):
            return match.groupdict()

    def parse(self, parsable: typing.Dict[str, str]):
        """Parse a line that reads config-file input."""
        parsed = {}
        key = parsable['key'].strip(' "')
        mode = parsable['mode']
        args = parsable['args'].split(',')
        parsed.update(self._normalize(mode, args))
        return key, parsed

    _keys = {
        'readBool': ['defaultVal'],
        'readInt': ['defaultVal', 'minVal', 'maxVal'],
        'readDouble': ['defaultVal', 'minVal', 'maxVal'],
        'readString': ['defaultVal'],
        'readDoubleArray': [
            'defaultSize',
            'size',
            'defaultVal',
            'minVal',
            'maxVal',
        ],
    }

    def _normalize(self, mode: str, args: typing.Iterable[str]):
        """Determine argument type and keys from `mode`."""
        pairs = dict(zip(self._keys[mode], [arg.strip(' "') for arg in args]))
        return {'mode': mode, **pairs}


class VariableDeclaration:
    """Pattern parser for variable declarations."""

    @property
    def pattern(self):
        return re.compile(r"""
            # the start of the string, followed by optional whitespace
            \A\s*
            # type declaration
            Scalar\_t
            # optional whitespace
            \s*
            # variable name with dereference operator
            \*(?P<name>\w+)
            # C statement terminator
            \;
            # optional whitespace, followed by the end of the string
            \s*\Z
        """, re.VERBOSE)

    def match(self, line: str):
        """Identify lines that define a variable."""
        if match := self.pattern.match(line.strip()):
            return match.groupdict()

    def parse(self, parsable: typing.Dict[str, str]):
        """Parse information about a defined variable."""
        parsed = {
            'size': None,
            'value': None,
        }
        return parsable['name'], parsed


class VariableDefinition:
    """Pattern parser for variable definitions."""

    # TODO:
    # - __init__ should take arguments that allow the user to define the
    #   variable type (e.g., 'Scalar_t').
    # - Can the pattern identify things like `size` and `value` with optional
    #   user input, or does `parse` need to handle that?

    @property
    def pattern(self):
        return re.compile(r"""
            # the start of the string, followed by optional whitespace
            \A\s*
            # type declaration
            Scalar\_t
            # optional whitespace
            \s*
            # variable name
            (?P<name>\w+)
            # array size
            \[(?P<size>\d+)\]
            # equals sign surrounded by optional whitespace
            \s*=\s*
            # array value(s)
            \{(?P<value>(\d*(?:\.\d+)?)|((?:config\.)?\w+))\}
            # C statement terminator
            \;
            # optional whitespace, followed by the end of the string
            \s*\Z
        """, re.VERBOSE)

    def match(self, line: str):
        """Identify lines that define a variable."""
        if match := self.pattern.match(line.strip()):
            return match.groupdict()

    def parse(self, parsable: typing.Dict[str, str]):
        """Parse information about a defined variable."""
        parsed = {
            'size': int(parsable['size']),
            'value': f"[{','.join(container.wrap(parsable['value']))}]",
        }
        return parsable['name'], parsed


class ConfigurationC(SourceFile):
    """A representation of default arguments in EPREM `configuration.c`."""

    _types = container.Bijection(
        {
            'readBool': bool,
            'readInt': int,
            'readDouble': float,
            'readString': str,
            'readDoubleArray': list,
        }
    )

    def __init__(self, srcdir: typing.Union[str, pathlib.Path]=None) -> None:
        super().__init__('CONFIGURATION_C', 'configuration.c', srcdir)

    def standardize(self, loaded: dict):
        return self._replace(loaded, 'mode', 'type', self._types)

    def load_from_source(self, file: paths.TextFile) -> dict:
        assignments = self._get_assignments(file)
        arrays = self._get_array_defaults(file)
        subs = {
            key: {
                'mode': assigned['mode'],
                'defaultVal': arrays[assigned['defaultVal']]['value'],
                'maxVal': assigned['maxVal'],
                'minVal': assigned['minVal'],
            } for key, assigned in assignments.items()
            if assigned['defaultVal'] in arrays
        }
        return {
            key: subs.get(key, attrs)
            for key, attrs in assignments.items()
        }

    def __getitem__(self, key: str) -> typing.Dict[str, typing.Any]:
        """Request a reference object by parameter name."""
        if key in self.definitions:
            return self.definitions[key]
        raise KeyError(f"No reference information for {key!r}")

    def get(self, key: str, default: typing.Any=None, format: str=None):
        if format == 'raw':
            modes = self._types.invert()
            loaded = self._replace(self.definitions, 'type', 'mode', modes)
            return loaded.get(key, default)
        if format == 'full':
            defined = self.definitions.get(key, default)
            loaded = {**defined, **_metadata._CONFIGURATION_C.get(key, {})}
            return {
                k: v.__name__ if k == 'type' else v
                for k, v in loaded.items()
            }
        return super().get(key, default, format)

    def tostring(self, **kwargs):
        s = super().tostring(**kwargs)
        return 'CONFIGURATION_C = {\n' + f"{s}" + '\n}'

    def _get_assignments(self, file: paths.TextFile):
        """Get the assigned default values from function calls."""
        pattern = FunctionCall()
        return file.extract(pattern.match, pattern.parse)

    def _get_array_defaults(self, file: paths.TextFile):
        """Get all array default values defined in the source."""
        patterns = (VariableDeclaration(), VariableDefinition())
        defaults = {}
        for pattern in patterns:
            current = file.extract(pattern.match, pattern.parse)
            defaults.update(current)
        return defaults

    def _replace(
        self,
        mapping: typing.Mapping[str, typing.Mapping],
        old: str,
        new: str,
        conversion: typing.Mapping,
    ) -> typing.Mapping[str, typing.Mapping]:
        """Convert values in the interior mappings of `mapping`."""
        return {
            key: {
                new: conversion[interior[old]],
                **{k: v for k, v in interior.items() if k != old}
            } for key, interior in mapping.items()
        }





"""
Operators on arbitrary numeric objects.

This module provides a canonical namespace for all relevant standard-library
numeric operators. It also defines the aliased mappings that appear as
`~numeric.OPERATIONS` and `~numeric.OPERATORS`, as well as a utility function,
`get`, for retrieving canonical operators by name or functional group.
"""

from builtins import (
    pow,
    round,
)
from math import (
    floor,
    ceil,
    trunc,
)
from operator import (
    eq,
    ne,
    lt,
    le,
    gt,
    ge,
    abs,
    pos,
    neg,
    add,
    sub,
    mul,
    truediv,
    floordiv,
    mod,
)
import typing


__all__ = [
    'eq',
    'ne',
    'lt',
    'le',
    'gt',
    'ge',
    'abs',
    'pos',
    'neg',
    'round',
    'floor',
    'ceil',
    'trunc',
    'add',
    'sub',
    'mul',
    'truediv',
    'floordiv',
    'mod',
    'pow',
]


def __int__(self):
    """Called for int(self)."""
    return int(self)


def __float__(self):
    """Called for float(self)."""
    return float(self)


def __complex__(self):
    """Called for complex(self)."""
    return complex(self)


def __eq__(self, other):
    """Called for self == other."""
    return eq(self, other)


def __ne__(self, other):
    """Called for self != other."""
    return ne(self, other)


def __lt__(self, other):
    """Called for self < other."""
    return lt(self, other)


def __gt__(self, other):
    """Called for self > other."""
    return gt(self, other)


def __le__(self, other):
    """Called for self <= other."""
    return le(self, other)


def __ge__(self, other):
    """Called for self >= other."""
    return ge(self, other)


def __abs__(self):
    """Called for abs(self)."""
    return abs(self)


def __pos__(self):
    """Called for +self."""
    return pos(self)


def __neg__(self):
    """Called for -self."""
    return neg(self)


def __round__(self, ndigits: typing.Optional[int]=None):
    """Called for round(self)."""
    return round(self, ndigits=ndigits)


def __floor__(self):
    """Called for math.floor(self)."""
    return floor(self)


def __ceil__(self):
    """Called for math.ceil(self)."""
    return ceil(self)


def __trunc__(self):
    """Called for math.trunc(self)."""
    return trunc(self)


def __add__(self, other):
    """Called for self + other."""
    return add(self, other)


def __radd__(self, other):
    """Called for other + self."""
    return add(other, self)


def __iadd__(self, other):
    """Called for self += other."""
    return add(self, other)


def __sub__(self, other):
    """Called for self - other."""
    return sub(self, other)


def __rsub__(self, other):
    """Called for other - self."""
    return sub(other, self)


def __isub__(self, other):
    """Called for self -= other."""
    return sub(self, other)


def __mul__(self, other):
    """Called for self * other."""
    return mul(self, other)


def __rmul__(self, other):
    """Called for other * self."""
    return mul(other, self)


def __imul__(self, other):
    """Called for self *= other."""
    return mul(self, other)


def __truediv__(self, other):
    """Called for self / other."""
    return truediv(self, other)


def __rtruediv__(self, other):
    """Called for other / self."""
    return truediv(other, self)


def __itruediv__(self, other):
    """Called for self /= other."""
    return truediv(self, other)


def __floordiv__(self, other):
    """Called for self // other."""
    return floordiv(self, other)


def __rfloordiv__(self, other):
    """Called for other // self."""
    return floordiv(other, self)


def __ifloordiv__(self, other):
    """Called for self //= other."""
    return floordiv(self, other)


def __mod__(self, other):
    """Called for self % other."""
    return mod(self, other)


def __rmod__(self, other):
    """Called for other % self."""
    return mod(other, self)


def __imod__(self, other):
    """Called for self %= other."""
    return mod(self, other)


def __pow__(self, other, mod: typing.Optional[int]=None):
    """Called for self ** other."""
    return pow(self, other, mod=mod)


def __rpow__(self, other, mod: typing.Optional[int]=None):
    """Called for other ** self."""
    return pow(other, self, mod=mod)


def __ipow__(self, other, mod: typing.Optional[int]=None):
    """Called for self **= other."""
    return pow(self, other, mod=mod)





import numbers
import typing

import numpy
import numpy.typing


R = typing.TypeVar('R', bound=numbers.Real)

class Member:
    """An arbitrary object from a dataset."""

    def __init__(self, **attributes) -> None:
        self._attributes = attributes
        self._strings = {}

    def __repr__(self) -> str:
        """An unambiguous representation of this object."""
        these = {
            k: self._strings.get(k, v)
            for k, v in self._attributes.items()
        }
        attrs = ', '.join(f"{k}={v!r}" for k, v in these.items())
        return f"{self.__class__.__qualname__}({attrs})"


class Array(Member):
    """An array-like dataset object."""

    def __init__(
        self,
        data: numpy.typing.ArrayLike,
        unit: str,
        dimensions: typing.Tuple[str],
    ) -> None:
        super().__init__(data=data, unit=unit, dimensions=dimensions)
        self._strings['data'] = type(data)

    def __getitem__(self, index, /):
        """Called for self[index]."""
        return self.data[index]

    def __array__(self, *args, **kwargs):
        """Called for conversion to a `numpy.ndarray`."""
        array = numpy.asarray(self.data)
        return numpy.array(array, *args, **kwargs)

    @property
    def data(self) -> numpy.typing.ArrayLike:
        """The array-like data interface."""
        return self._attributes['data']

    @property
    def unit(self) -> str:
        """The metric unit of this array's data values."""
        return self._attributes['unit']

    @property
    def dimensions(self) -> typing.Tuple[str]:
        """The names of this array's axes."""
        return self._attributes['dimensions']


class Scalar(Member):
    """A scalar dataset object."""

    def __init__(self, value: R, unit: str) -> None:
        super().__init__(value=value, unit=unit)

    def __float__(self) -> float:
        """Called for float(self)."""
        return float(self.value)

    def __int__(self) -> int:
        """Called for int(self)."""
        return int(self.value)

    @property
    def value(self) -> R:
        """The value of this scalar quantity."""
        return self._attributes['value']

    @property
    def unit(self) -> str:
        """The metric unit of this scalar quantity's value."""
        return self._attributes['unit']


class Axis(Member):
    """A dataset array axis."""

    def __init__(self, size: int) -> None:
        super().__init__(size=size)

    @property
    def size(self) -> int:
        """The size of this axis."""
        return self._attributes['size']


def array_factory(
    data: numpy.typing.ArrayLike,
    unit: str,
    dimensions: typing.Iterable[str],
) -> Array:
    """Create an object representing a dataset array."""
    return Array(data, unit, tuple(dimensions))


def scalar_factory(value: R, unit: str):
    """Create an object representing a dataset scalar."""
    return Scalar(value, unit)


def axis_factory(size: int):
    """Create an object representing a dataset axis."""
    return Axis(size)





import numbers
import typing

import numpy
import numpy.typing

from .. import measured
from ._axis import (
    Points,
    Symbols,
    Coordinates,
)


@Points.register.factory
def points_factory(x, /):
    """Internal factory for `~Points` instances."""
    return Points(points_args(x))


def points_args(x, /) -> numpy.typing.NDArray[numpy.uint]:
    """Parse arguments to initialize `~Points`."""
    if isinstance(x, Points):
        return x.data
    try:
        # XXX: Creating this intermediate array circumvents bugs that arise when
        # `x.__array__` does not accept arguments.
        tmp = numpy.array(x)
    except (ValueError, TypeError) as err:
        raise TypeError(
            f"Cannot initialize {Points} from {x}"
        ) from err
    data = numpy.array(tmp, ndmin=1, dtype=int)
    if any(i < 0 for i in data):
        raise TypeError(
            f"Cannot initialize {Points} from an object"
            " with negative values"
        ) from None
    return data


@Symbols.register.factory
def symbols_factory(x, /):
    """Internal factory for `~Symbols` instances."""
    return Symbols(symbols_args(x))


def symbols_args(x, /) -> typing.List[str]:
    """Parse arguments to initialize `~Symbols`."""
    if isinstance(x, Symbols):
        return x.data
    if isinstance(x, str):
        return [x]
    if not isinstance(x, typing.Sequence):
        raise TypeError(
            f"Cannot initialize {Symbols} from non-sequence"
            f" type {type(x)}"
        ) from None
    if not all(isinstance(i, str) for i in x):
        raise TypeError(
            f"Cannot initialize {Symbols} from sequence"
            " with non-string members"
        ) from None
    return x


@Coordinates.register.factory
def coordinates_factory(x, /):
    """Internal factory for `~Coordinates` instances."""
    return Coordinates(coordinates_args(x))


def coordinates_args(x, /) -> measured.Sequence[numbers.Real]:
    """Parse arguments to initialize `~Coordinates`."""
    if isinstance(x, Coordinates):
        return x.data
    try:
        s = measured.sequence(x)
    except TypeError as err:
        raise TypeError(
            f"Cannot initialize {Coordinates} from {x}"
        ) from err
    return s





"""
Support for the package-wide command-line interface (CLI).
"""

import datetime
import os
import typing

from . import etc
from . import measured
from . import parameter
from . import paths


class UserError(Exception):
    """Incorrect use of a function (probably from the CLI)."""


class CLIError(Exception):
    """Invalid data generated by the CLI."""


def run(mode: str, options: typing.Mapping):
    if mode == 'configfile':
        return run_configfile_subparser(options)
    raise CLIError(f"Unknown mode: {mode!r}")


def run_configfile_subparser(options: typing.Mapping):
    """Generate or compare configuration files."""
    if options.get('generate'):
        return generate_config_file(
            path=options['generate'],
            source=options['source'],
            verbose=options['verbose'],
        )
    if options.get('files'):
        return compare_config_files(
            files=options['files'],
            source=options['source'],
            diff=options['diff'],
            names=options['names'],
        )


_timestr = datetime.datetime.now().strftime('%Y-%m-%d %H:%M:%S')


def generate_config_file(
    path: paths.PathLike,
    source: paths.PathLike=None,
    verbose: bool=False,
) -> None:
    """Create a configuration file with default parameter values."""
    default = parameter.interface(srcdir=source)
    keys = sorted(parameter.ConfigurationC(source))
    header = [
        f"# EPREM default configuration file\n",
        f"# Generated at {_timestr}\n",
        f"# --------------------------------\n",
    ]
    values = [
        f"{key}={_get_arg_value(default[key])}\n"
        for key in keys
    ]
    footer = [
        f"# --------------------------------\n",
    ]
    filepath = paths.fullpath(path)
    with filepath.open('w') as fp:
        fp.writelines(header)
        fp.writelines(values)
        fp.writelines(footer)
    if verbose:
        print(f"Wrote default config file to {filepath}")


def compare_config_files(
    files: typing.Iterable[paths.PathLike],
    source: paths.PathLike=None,
    diff: bool=False,
    names: bool=False,
) -> None:
    """Compare values of EPREM configuration-file parameters.
    
    This method will print the name of each parameter and its default value, as
    well as the corresponding value contained in each configuration file in
    `files`. If `source` is given, it will read parameter names and default
    values from the version of `configuration.c` in `source`; otherwise it will
    use the values in the local database.
    """
    if not files:
        raise UserError("Nothing to compare") from None
    args = _build_arg_dict(files, source=source, diff=diff, names=names)
    topkeys = next(list(v.keys()) for v in args.values() if v)
    nonnull = (v for item in args.values() for v in item.values() if v)
    # width (i.e., string length) of the longest parameter
    pwidth = max(len(k) for k in args)
    # width (i.e., string length) of the longest file key
    lwidth = max(len(k) for k in topkeys)
    # width (i.e., string length) of the longest non-null value
    vwidth = max(len(v) for v in nonnull)
    # padded parameter-value width
    rwidth = vwidth + 2
    # amount of parameter-value right-justification
    jwidth = min(78, rwidth)
    # total width: either the longest parameter or the longest key-value combo.
    # - longest parameter produces groups like
    # * ParameterName
    # * -------------
    # * key     value
    # - longest key-value combo produces groups like
    # *    ParameterName
    # * -------------------
    # * key           value
    cwidth = max(pwidth, lwidth + jwidth)
    print()
    for key, item in args.items():
        print(str(key).center(cwidth))
        print('-' * cwidth)
        for k, v in item.items():
            print(f"{str(k).ljust(lwidth)}{str(v).rjust(cwidth-lwidth)}")
        print()


def _build_arg_dict(
    files: typing.Iterable[paths.PathLike],
    source: paths.PathLike=None,
    diff: bool=False,
    names: bool=False,
) -> typing.Dict[str, typing.Dict[str, typing.Any]]:
    """Build a `dict` of parameter values."""
    default = parameter.interface()
    normalized = _normalize_paths(files)
    if names:
        mapped = normalized
    else:
        mapped = {k: str(i) for i, k in enumerate(normalized)}
    targets = {
        relpath: parameter.interface(abspath)
        for abspath, relpath in mapped.items()
    }
    keys = sorted(parameter.ConfigurationC(source))
    built = {key: {} for key in keys}
    for key in keys:
        heading = f"  {key} (default: {_format_arg_string(default[key])})  "
        current = _map_parameter(key, targets)
        values = list(current.values())
        if not diff or _any_not_equal(values, default[key]):
            built[heading] = current.copy()
    if diff:
        return {k: v for k, v in built.items() if v}
    return built


def _map_parameter(
    key: str,
    targets: typing.Dict[str, parameter.Interface]
) -> typing.Dict[str, str]:
    """Build a dict of values for a single parameter."""
    d = {}
    for name, interface in targets.items():
        userdef = parameter.configfile(interface.config)
        if userdef.get(key):
            d[name] = _format_arg_string(interface.get(key))
        else:
            d[name] = '(default)'
    return d


def _any_not_equal(values: list, default):
    """Check for unequal values, excluding default flag."""
    replaced = values.copy()
    for i, vi in enumerate(values):
        if vi == '(default)':
            replaced[i] = _format_arg_string(default)
    v0 = replaced[0]
    return any(vi != v0 for vi in replaced)


def _normalize_paths(args: typing.Iterable[paths.PathLike]):
    """Compute full and unique relative paths."""
    full = {
        str(arg): str(paths.fullpath(arg, strict=True))
        for arg in args
    }
    common = os.path.commonpath(full.values())
    return {
        str(path): os.path.relpath(path, start=common)
        for path in full.values()
    }


def _format_arg_string(arg):
    """Format this argument for display."""
    if isinstance(arg, measured.Object):
        data = _get_arg_value(arg)
        if arg.unit == '1' or data is None:
            return f"{data!r}"
        if str(arg.unit) in {'d', 'day'}:
            return f"{data} 'day'"
        return f"{data} {str(arg.unit)!r}"
    return str(arg)


def _get_arg_value(arg):
    """Extract this argument's value, if necessary."""
    if isinstance(arg, measured.Object):
        return arg.data[0] if len(arg.data) == 1 else arg.data
    return arg


_docstring_replacements = {
    '`src`': 'SRC',
    '`source`': 'SRC',
    "`paths`": "the given paths",
    '`': "'",
    "the local database": str(parameter.SourceFile._db_path),
}

DESCRIPTIONS = {
    'configfile': etc.doc2help(
        run_configfile_subparser,
        mode='full',
        replacements=_docstring_replacements,
    ),
}




"""
Support for objects that manage numeric data.
"""

import numbers
import typing

import numpy.typing

from . import _data as data
from . import _operators as operators
from . import _implementation as implementation
from . import _index as index
from . import _factory as factory
from ._array import (
    Array,
    array_factory as array,
    Dimensions,
    dimensions_factory as dimensions,
    remesh,
)
from ._operations import (
    _OPERATIONS as OPERATIONS,
    _OPERATORS as OPERATORS,
    IMPLEMENTATIONS,
    FUNCTIONS,
    METHODS,
    DUNDER,
    CALLABLES,
    OperandTypeError,
)
from ._mixins import (
    Functions,
    Operators,
)
from ._objects import (
    Quantity,
    Interface,
    Object,
    Type,
)

# HACK: This should be in the public namespace.
from ._implementation import Result

# HACK: Define this to avoid import errors until full deprecation.
Implementation = object


__all__ = [
    'array',
    'data',
    'dimensions',
    'implementation',
    'index',
    'operators',
    'remesh',
    'factory',
    'Functions',
    'Operators',
    'Array',
    'Dimensions',
    'Quantity',
    'Interface',
    'Object',
    'OperandTypeError',
    'Result',
    'Type',
    'CALLABLES',
    'DUNDER',
    'FUNCTIONS',
    'IMPLEMENTATIONS',
    'METHODS',
    'OPERATIONS',
    'OPERATORS',
]


@typing.runtime_checkable
class ArrayType(typing.Protocol):
    """Structural protocol for numeric arrays."""

    @property
    def dimensions(self) -> Dimensions: ...

    @property
    def array(self) -> numpy.typing.NDArray: ...

    @property
    def shapemap(self) -> typing.Dict[str, int]: ...

    @property
    def shape(self) -> typing.Tuple[int, ...]: ...


ArrayLike = typing.Union[
    numpy.typing.NDArray[numpy.number],
    typing.Sequence[numbers.Number],
]
"""Type alias for array-like numeric objects."""
# TODO: Include nested sequences.





"""
Interfaces to runtime parameter values.
"""

import pathlib
import typing

import numpy

from .. import aliased
from .. import container
from .. import measured
from .. import metric
from .. import numeric
from .. import paths
from .. import physical
from .. import real
from . import _config
from . import _metadata


T = typing.TypeVar('T')


class Variable(physical.Vector[real.ValueType]):
    """A vector-like object with conditional cast to built-in type.

    An instance of this class is a physical vector that also supports converting
    single-valued instances to built-in `int`, `float`, or `complex` type.
    """

    def __getitem__(self, *args):
        """Extract a sub-collection.
        
        Notes
        -----
        - This method is designed to return a `~Scalar` when the given arguments
          produce 0-D data.
        """
        idx = container.unwrap(args)
        if isinstance(idx, typing.SupportsIndex) and idx < 0:
            idx += len(self)
        data = self.data[idx]
        if container.length(data) == 0:
            return physical.scalar(data, unit=self.unit)
        return variable_factory(data, unit=self.unit)

    def __int__(self):
        return self._as_builtin(int)

    def __float__(self):
        return self._as_builtin(float)

    def __complex__(self):
        return complex(float(self))

    def __round__(self, ndigits=None):
        """Called for round(self)."""
        return self._as_scalar(numeric.operators.round, ndigits=ndigits)

    def __floor__(self):
        """Called for math.floor(self)."""
        return self._as_scalar(numeric.operators.floor)

    def __ceil__(self):
        """Called for math.ceil(self)."""
        return self._as_scalar(numeric.operators.ceil)

    def __trunc__(self):
        """Called for math.trunc(self)."""
        return self._as_scalar(numeric.operators.trunc)

    def _as_scalar(self, f: typing.Callable, **kwargs):
        """Return the result of `f` as a `~Scalar`."""
        try:
            data = f(self.data[0], **kwargs)
        except TypeError as err:
            raise TypeError(
                f"Cannot apply {f.__name__} to size-{self.size} variable"
            ) from err
        return physical.scalar(data, self.unit)

    def _as_builtin(self, t: T):
        """Convert this instance to a built-in numeric type."""
        try:
            result = t(self.data[0])
        except TypeError as err:
            raise TypeError(
                f"Cannot convert a size-{self.size} variable to {t!r}"
            ) from err
        return result

    def _singular(self):
        """Represent this object as a single value only if possible."""
        size = self.size
        if size == 1:
            return self.data[0]
        raise TypeError(
            f"Cannot represent a size-{size} variable as a number"
        ) from None


@Variable.register.factory
def variable_factory(x, /, unit=None) -> Variable:
    """Factory function for `~Variable`."""
    d, u = variable_args(x, unit)
    ctx = measured.context(u, measured.convert.arraylike)
    return Variable(d, ctx)


def variable_args(x, unit, /):
    """Parse arguments to initialize `~Variable`.

    Notes
    -----
    - If `x` is an instance of `~Variable`, this function will return copies of
      the data and unit. Otherwise, it will attempt to construct an instance of
      `~physical.Vector` from the data and unit, then return a copy of that
      vector's data and unit.
    """
    if isinstance(x, measured.Object):
        vector = physical.vector(x)
        return vector.data, vector.unit
    try:
        vector = physical.vector(numpy.array(x, ndmin=1), unit=unit)
    except (TypeError, ValueError) as err:
        lowercase = ('vector', 'variable')
        capitalized = ('Vector', 'Variable')
        m = str(err).replace(*lowercase).replace(*capitalized)
        raise type(err)(m)
    return vector.data, vector.unit


_VT = typing.TypeVar('_VT')


class Interface(aliased.Mapping[str, _VT]):
    """An interface to EPREM parameter values."""

    def __init__(
        self,
        base: typing.Mapping[str, typing.Any],
        defaults: _config.Defaults,
        cfgpath: typing.Optional[pathlib.Path],
    ) -> None:
        """Interface initializer.

        Not meant for public use. Please see `~interface`.
        """
        super().__init__(base, aliases=_metadata.ALIASES)
        self._defaults = defaults
        self._config = cfgpath

    def __repr__(self) -> str:
        clsname = self.__class__.__qualname__
        if self.config is None:
            return f"{clsname}(source={str(self.source)!r})"
        display = {
            'config': str(self.config),
            'source': str(self.source),
        }
        args = ', '.join(f"{k}={v!r}" for k, v in display.items())
        return f"{clsname}({args})"

    def __getitem__(self, __k: str):
        """Create the appropriate object for the named parameter."""
        try:
            this = super().__getitem__(__k)
        except KeyError:
            raise KeyError(f"No parameter corresponding to {__k!r}") from None
        value = this['value']
        unit = this['unit']
        usedeg = super().__getitem__('useDegrees')
        if unit == 'rad' and usedeg['value']:
            unit = 'deg'
        if unit is None:
            return value
        return variable_factory(value, unit=unit)

    @property
    def source(self):
        """The source of default values.
        
        If this interface was initialized with a path to an EPREM `src`
        directory, the value of this property will be that path. Otherwise, it
        will be the path to the database of default values.
        """
        return self.reference.origin

    @property
    def basetypes(self):
        """Values of constants defined in `baseTypes.h`."""
        return self._defaults.basetypes

    @property
    def reference(self):
        """Metadata and default values from from `configuration.c`."""
        return self._defaults.reference

    @property
    def config(self):
        """The full path to the relevant configuration file, if given."""
        return self._config


def interface_factory(*args, **kwargs):
    cfginit, srcinit, options = _init_interface(args, kwargs)
    defaults = _config.Defaults(srcdir=srcinit)
    local = {
        key: info.get('default')
        for key, info in _metadata._LOCAL.items()
    }
    init = {**local, **defaults}
    cfgpath = None
    if cfginit:
        cfgfile = _config.configfile_factory(cfginit, **options)
        cfgpath = cfgfile.source
        user = {
            # NOTE: this does not use `configfile.get(key)` because we don't
            # want to overwrite the default value with `None` if the config
            # file doesn't include a given parameter.
            key: _config._convert(cfgfile[key], parameter['type'])
            for key, parameter in defaults.reference.items()
            if key in cfgfile
        }
        init.update(user)
    base = {
        key: {'unit': _metadata.UNITS.get(key), 'value': value}
        for key, value in init.items()
    }
    return Interface(base, defaults, cfgpath)


def _init_interface(
    args: typing.Tuple[paths.PathLike, paths.PathLike],
    kwargs: dict,
) -> typing.Tuple[paths.PathLike, paths.PathLike, dict]:
    """Internal initialization parser for `~interface` function."""
    if len(args) == 2:
        return *args, kwargs
    if len(args) == 1:
        source = kwargs.pop('srcdir', None)
        return args[0], source, kwargs
    if len(args) == 0:
        config = kwargs.pop('config', None)
        source = kwargs.pop('srcdir', None)
        return config, source, kwargs
    raise TypeError(args, kwargs)





"""
Public namespace for ABCs corresponding to protocol classes.
"""

from ._abcs import (
    Ordered,
    Comparable,
    Additive,
    Multiplicative,
    Algebraic,
    Real,
    Scalar,
    Sequence,
    Array,
)


__all__ = [
    'Ordered',
    'Comparable',
    'Additive',
    'Multiplicative',
    'Algebraic',
    'Real',
    'Scalar',
    'Sequence',
    'Array',
]





"""
Values and metadata of universal physical constants.
"""

import collections.abc
import typing

import numpy

from . import aliased
from . import measured
from . import physical
from . import real


_constants = {
    'pi': {
        'info': "The ratio of a circle's circumference to its diameter.",
        'all': numpy.pi,
        'quantity': 'number',
    },
    'k': {
        'info': "Boltzmann's constant.",
        'mks': {'unit': 'J / K', 'value': 1.3807e-23},
        'cgs': {'unit': 'erg / K', 'value': 1.3807e-16},
        'quantity': 'energy / temperature'
    },
    'e': {
        'info': "Elementary charge.",
        'mks': {'unit': 'C', 'value': 1.6022e-19},
        'cgs': {'unit': 'statC', 'value': 4.8032e-10},
        'quantity': 'charge',
    },
    'me': {
        'info': "Electron mass.",
        'mks': {'unit': 'kg', 'value': 9.1094e-31},
        'cgs': {'unit': 'g', 'value': 9.1094e-28},
        'quantity': 'mass',
    },
    'mp': {
        'info': "Proton mass.",
        'mks': {'unit': 'kg', 'value': 1.6726e-27},
        'cgs': {'unit': 'g', 'value': 1.6726e-24},
        'quantity': 'mass',
    },
    'G': {
        'info': "Gravitational constant.",
        'mks': {'unit': 'm^3 / (s^2 * kg)', 'value': 6.6726e-11},
        'cgs': {'unit': 'dyn * cm^2 / g^2', 'value': 6.6726e-8},
        'quantity': 'force * area / mass^2',
    },
    'g': {
        'info': "Gravitational acceleration.",
        'mks': {'unit': 'm / s^2', 'value': 9.8067},
        'cgs': {'unit': 'cm / s^2', 'value': 9.8067e2},
        'quantity': 'acceleration',
    },
    'h': {
        'info': "Planck's constant.",
        'mks': {'unit': 'J * s', 'value': 6.6261e-34},
        'cgs': {'unit': 'erg * s', 'value': 6.6261e-27},
        'quantity': 'energy * time',
    },
    'c': {
        'info': "Speed of light in a vacuum.",
        'mks': {'unit': 'm / s', 'value': 2.99792458e8},
        'cgs': {'unit': 'cm / s', 'value': 2.99792458e10},
        'quantity': 'speed',
    },
    'epsilon0': {
        'info': "Permittivity of free space.",
        'mks': {'unit': 'F / m', 'value': 8.8542e-12},
        'cgs': {'unit': '1', 'value': 1.0},
        'quantity': 'permittivity',
    },
    'mu0': {
        'info': "Permeability of free space.",
        'mks': {'unit': 'H / m', 'value': 4*numpy.pi * 1e-7},
        'cgs': {'unit': '1', 'value': 1.0},
        'quantity': 'permeability',
    },
    'Rinfinity': {
        'info': "Rydberg constant.",
        'mks': {'unit': '1 / m', 'value': 1.0974e7},
        'cgs': {'unit': '1 / cm', 'value': 1.0974e5},
        'quantity': '1 / length',
    },
    'a0': {
        'info': "Bohr radius.",
        'mks': {'unit': 'm', 'value': 5.2918e-11},
        'cgs': {'unit': 'cm', 'value': 5.2918e-9},
        'quantity': 'length',
    },
    're': {
        'info': "Classical electron radius.",
        'mks': {'unit': 'm', 'value': 2.8179e-15},
        'cgs': {'unit': 'cm', 'value': 2.8179e-13},
        'quantity': 'length',
    },
    'alpha': {
        'info': "Fine structure constant.",
        'mks': {'unit': '1', 'value': 7.2974e-3},
        'cgs': {'unit': '1', 'value': 7.2974e-3},
        'quantity': 'number',
    },
    'c1': {
        'info': "First radiation constant.",
        'mks': {'unit': 'W * m^2', 'value': 3.7418e-16},
        'cgs': {'unit': 'erg * cm^2 / s', 'value': 3.7418e-16},
        'quantity': 'power * area',
    },
    'c2': {
        'info': "Second radiation constant.",
        'mks': {'unit': 'm * K', 'value': 1.4388e-2},
        'cgs': {'unit': 'cm * K', 'value': 1.4388},
        'quantity': 'length * temperature',
    },
    'sigma': {
        'info': "Stefan-Boltzmann constant.",
        'mks': {'unit': 'W / (m^2 * K^4)', 'value': 5.6705e-8},
        'cgs': {'unit': '(erg / s) / (cm^2 * K^4)', 'value': 5.6705e-5},
        'quantity': 'power / (area * temperature^4)',
    },
    'eV': {
        'info': "Energy associated with 1 eV.",
        'mks': {'unit': 'J', 'value': 1.6022e-19},
        'cgs': {'unit': 'erg', 'value': 1.6022e-12},
        'quantity': 'energy',
    },
    'amu': {
        'info': "Atomic mass unit.",
        'mks': {'unit': 'kg', 'value': 1.6605e-27},
        'cgs': {'unit': 'g', 'value': 1.6605e-24},
        'quantity': 'mass',
    },
    'au': {
        'info': "Astronomical unit.",
        'mks': {'unit': 'm', 'value': 1.495978707e11},
        'cgs': {'unit': 'cm', 'value': 1.495978707e13},
        'quantity': 'length',
    },
}
_constants['H+'] = {
    'info': "First ionization energy of hydrogen.",
    **{
        k: {
            'unit': _constants['eV'][k]['unit'],
            'value': 13.6 * _constants['eV'][k]['value'],
        }
        for k in ('mks', 'cgs')
    },
    'quantity': 'energy',
}
_constants['MeV'] = {
    'info': _constants['eV']['info'].replace('eV', 'MeV'),
    **{
        k: {
            'unit': _constants['eV'][k]['unit'],
            'value': 1e6 * _constants['eV'][k]['value'],
        }
        for k in ('mks', 'cgs')
    },
    'quantity': 'energy',
}


def _normalize(
    defined: typing.Dict[str, typing.Dict[str, typing.Any]],
) -> typing.Dict[str, typing.Dict[str, typing.Any]]:
    """Normalize constant definitions."""
    systems = ('mks', 'cgs')
    norm = {}
    for key, definition in defined.items():
        if 'all' in definition:
            value = definition.pop('all')
            new = {k: {'unit': '1', 'value': value} for k in systems}
            definition.update(new)
        norm[key] = definition
    return norm


# NOTE: I'm not sure that this needs to be an aliased mapping.
CONSTANTS = aliased.Mapping(_normalize(_constants))


class Constant(measured.Value[real.ValueType]):
    """A universal physical constant.
    
    An instance of this class represents the value of a named physical constant
    in a particular metric system, as well as the following metadata

    - the system-independent metric quantity
    - the system-specific metric unit, if applicable
    - a brief description
    """

    def __init__(
        self,
        value: measured.Value,
        /,
        quantity: str,
        info: str,
    ) -> None:
        """Called to initialize a constant object.
        
        It is usually not necessary to create instances of this class by hand.
        Please see `~Constants` for a mapping of universal physical constants
        with values and units in a given metric system.
        """
        super().__init__(value.data, value.context)
        self._quantity = quantity
        self._info = info

    @property
    def asscalar(self):
        """Extract the value and unit as a scalar object."""
        return physical.scalar(self.data, unit=self.unit)

    def __str__(self) -> str:
        base = super().__str__()
        extra = f"quantity={self.quantity!r}, info={self.info!r}"
        return f"{base}, {extra}"

    @property
    def quantity(self):
        """The corresponding metric quantity."""
        return self._quantity

    @property
    def info(self):
        """A brief description of this constant."""
        return self._info


class Constants(collections.abc.Mapping):
    """Universal physical constants in a given metric system."""
    def __init__(self, system: str) -> None:
        self.system = system.lower()
        self._mapping = CONSTANTS.copy()

    def __len__(self) -> int:
        """The number of defined constants."""
        return len(self._mapping)

    def __iter__(self) -> typing.Iterator:
        """Iterate over defined constants."""
        return iter(self._mapping)

    def __getitem__(self, name: str):
        """Create the named constant or raise an error."""
        try:
            definition = self._mapping[name]
        except KeyError as err:
            raise KeyError(f"Unknown constant: {name!r}") from err
        return self._create(definition)

    def _create(self, definition: dict):
        """Create a constant object from the given definition."""
        value = definition[self.system]['value']
        unit = definition[self.system]['unit']
        quantity = definition['quantity']
        info = definition['info']
        x = measured.value(value, unit=unit)
        return Constant(x, quantity, info)

    def __repr__(self) -> str:
        """An unambiguous representation of this object."""
        return f"{self.__class__.__qualname__}({self.system})"


MKS = Constants('mks')
"""Fundamental constants in MKS unit where applicable."""

CGS = Constants('cgs')
"""Fundamental constants in cgs unit where applicable."""


def show():
    """Print all defined physical constants."""
    for key, data in CONSTANTS.items(aliased=True):
        print(f"{key}: {data['info']}")
        for system in ('mks', 'cgs'):
            value = data[system]['value']
            unit = data[system]['unit']
            base = f"{system}: {value}"
            line = f"{base} [{unit}]" if unit != '1' else base
            print(f"\t{line}")
        print()


if __name__ == '__main__':
    show()





"""
Utility functions for checking data the property of numeric objects.
"""

import functools
import numbers
import typing

import numpy
import numpy.typing

from .. import container
from ._objects import Object


def isintegral(x):
    """True if is an object of integral type.

    This function exists to provide a single instance check against all integral
    types relevant to this package.
    """
    return isinstance(x, (numbers.Integral, numpy.integer))


def isclose(a: Object, b: numbers.Real) -> bool:
    """True if `b` is close to a value in `a`'s numeric data.
    
    This function is similar to (and, if fact, uses) `numpy.isclose`. The
    primary distinctions are that this function compares a single value to the
    real-valued data of a variable object and returns a single boolean value.
    
    Parameters
    ----------
    a : `~Object`
        The object that may contain `b`.

    b : real number
        The value for which to search.

    Notes
    -----
    - This function exists to handle cases in which floating-point arithmetic
      has caused a numeric operation to return an imprecise result, especially
      for small numbers (e.g., certain unit conversions). It will first test for
      standard containment via `in` before attempting to determine if `b` is
      close enough, within a very strict tolerance, to any member of `a`.
    """
    data = a.data
    if not container.isiterable(data):
        return a == b or numpy.isclose(b, data, atol=0.0)
    if b in data:
        return True
    if b < numpy.min(data) or b > numpy.max(data):
        return False
    return numpy.any([numpy.isclose(b, data, atol=0.0)])


def isequal(a, b):
    """True if `a` and `b` have equal numeric data.
    
    This is a convenience function that allows the caller to test whether two
    objects are numerically equivalent, even if they aren't strictly equal
    (perhaps because they have unequal metadata). If neither `a` nor `b` is an
    instance of `~Object`, calling this function is identical to calling
    `numpy.array_equal(a, b)`.
    """
    x = a.data if isinstance(a, Object) else a
    y = b.data if isinstance(b, Object) else b
    return numpy.array_equal(x, y)


@typing.overload
def hasdtype(
    a: typing.Union[Object, numpy.typing.ArrayLike],
    t: numpy.typing.DTypeLike,
    /,
): ...

@typing.overload
def hasdtype(
    a: typing.Union[Object, numpy.typing.ArrayLike],
    t: typing.Tuple[numpy.typing.DTypeLike],
    /,
): ...

def hasdtype(a, t, /):
    """True if `a` has one of the given data types.
    
    This function wraps `numpy.issubdtype` to allow the caller to pass more than
    one data type at a time, similar to the behavior of the built-in functions
    `isinstance` and `issubclass`.

    Parameters
    ----------
    a : `~Object` or array-like
        The object to check.
    t : dtype-like or tuple of dtype-like
        One or more objects that can be interpreted as `numpy` data types.

    Notes
    -----
    - If `a` is a `~Object`, this function will operate on `a.data`.
    - If the array-like operand (either `a` or `a.data`, as appropriate) is not
      a `numpy.ndarray`, this function will first convert it to one.
    """
    x = a.data if isinstance(a, Object) else a
    y = x if isinstance(x, numpy.ndarray) else numpy.array(x)
    dtype = y.dtype
    if isinstance(t, tuple):
        return any(numpy.issubdtype(dtype, i) for i in t)
    return numpy.issubdtype(dtype, t)


def implements(protocol, /):
    """Require that the data argument to a function implement `protocol`."""
    def wrapper(f):
        @functools.wraps(f)
        def check(self, data, *args, **kwargs):
            if isinstance(data, protocol):
                return f(self, data, *args, **kwargs)
            raise TypeError(
                f"Cannot instantiate class {type(self)}"
                f" with data that does not implement {protocol}"
            ) from None
        return check
    return wrapper





"""
Exception classes for the `metric` namespace.

Exceptions defined here:

- `UnitParsingError`
- `UnitConversionError`
- `SystemAmbiguityError`
- `UnitSystemError`
- `QuantityError`
"""


class UnitParsingError(Exception):
    """Error when attempting to parse string into unit."""

    def __init__(self, string: str) -> None:
        self.string = string

    def __str__(self) -> str:
        return f"Could not determine unit and magnitude of '{self.string}'"


class UnitConversionError(Exception):
    """Unknown unit conversion."""


class SystemAmbiguityError(Exception):
    """The metric system is ambiguous."""


class UnitSystemError(Exception):
    """The metric system does not contain this unit."""


class QuantityError(Exception):
    """An error occurred in `~metric.Quantity`."""





import functools
import typing


T = typing.TypeVar('T')
G = typing.TypeVar('G')


class Wrapper:
    """Substitute default values for exceptions.

    This class wraps a callable object in `try/except` logic that substitutes a
    default value when calling that object raises a known exception. The default
    behavior is to substitute `None`, but users may specify a specific
    substitution value when registering a known exception (see the `catch`
    method of this class).
    
    Examples
    --------
    Define a function that only inverts positive numbers:

    >>> def invert(v):
    ...     if v < 0:
    ...         raise ValueError
    ...     return 1 / v

    Create an instance of this class that wraps `invert`:

    >>> wrapper = exceptions.Wrapper(invert)

    Register `TypeError` with the default substitution, `None`, and register
    `ZeroDivisionError` with a custom substitution:
    
    >>> wrapper.catch(TypeError)
    >>> wrapper.catch(ZeroDivisionError, 'Bad!')

    A valid call to `invert` returns the expected result:

    >>> wrapper.call(2)
    0.5

    Passing a string argument raises a `TypeError` in `invert`, which `wrapper`
    converts to `None`:

    >>> wrapper.call('2')

    Triggering a `ZeroDivisionError` in `invert` causes `wrapper` to return the
    registered substitution:

    >>> wrapper.call(0)
    'Bad!'

    Passing a negative raises a `ValueError` in `invert`. Since we did not
    register this type of exception, it propagates back to us:

    >>> wrapper.call(-1)
    ...
    ValueError

    Notes
    -----
    This class was inspired by https://stackoverflow.com/a/8915613/4739101.
    """

    def __init__(self, __callable: typing.Callable[..., T]) -> None:
        self._f = __callable
        self._substitutions = {}

    def catch(self, exception: Exception, /, value: G=None):
        """Register a known exception and optional substitution value.
        
        Parameters
        ----------
        exception
            An exception class to catch when calling the wrapped object.

        value, optional
            The value to return from `call` when calling the wrapped object
            raises `exception`. The default value is `None`. There is one
            special case: Registering `exception` with `value = ...` (i.e., the
            built-in `Ellipsis` object), will cause `call` to return the given
            argument(s). See `call` for more information about the form of the
            return value in this case.
        """
        self._substitutions[exception] = value
        return self

    def compute(self, *args, **kwargs) -> typing.Union[T, G]:
        """Call the wrapped object with the given arguments.
        
        Parameters
        ----------
        *args
            Positional arguments to pass to the wrapped object.

        **kwargs
            Keyword arguments to pass to the wrapped object.

        Returns
        -------
        The result of calling the wrapped object, or an associated default
        value, or the given arguments.
        
        If no exceptions arose when calling the wrapped object, this method will
        return the result of that call.

        If calling the wrapped object raises a known exception and the value
        associated with that exception is not `...` (i.e., the built-in
        `Ellipsis` object), this method will return the associated value.
        
        If calling the wrapped object raises a known exception and the value
        associated with that exception is `...`, this method will return the
        given arguments. The return type in this case depends on the given
        arguments. If the user passes only a single positional argument, this
        method will return that argument. If the user passes only positional
        arguments, this method will return the equivalent `tuple`. If the user
        passes only keyword arguments, this method will return the equivalent
        `dict`. If the user passes positional and keyword arguments, this method
        will return a `tuple` containing the corresponding equivalent `tuple`
        and `dict`.
        
        If calling the wrapped object raises an exception that is unknown to
        this instance, that exception will propagate up to the caller as usual.
        """
        try:
            return self._f(*args, **kwargs)
        except tuple(self._substitutions) as err:
            value = self._substitutions[type(err)]
            if value != Ellipsis:
                return value
            if not kwargs:
                if len(args) == 1:
                    return args[0]
                return args
            if not args:
                return kwargs
            return args, kwargs


@typing.overload
def convert(expected, replacement, /):
    """Replace an expected exception."""

@typing.overload
def convert(expected: Exception, replacement: G, /) -> G:
    """Replace the expected exception with the given value.

    Parameters
    ----------
    expected : exception
        A subtype of `Exception` to catch.
    replacement
        The value to return after catching the expected exception.
    """

@typing.overload
def convert(
    expected: Exception,
    replacement: Exception,
    /,
    message: typing.Optional[str]=None,
) -> typing.NoReturn:
    """Replace the expected exception with a different exception.

    Parameters
    ----------
    expected : exception
        A subtype of `Exception` to catch.
    replacement : exception
        The exception to raise after catching the expected exception.
    message : string, optional
        The error message to include with the replacement exception.
    """

def convert(expected, replacement, /, message=None):
    def wrapper(f: typing.Callable[..., T]):
        @functools.wraps(f)
        def wrapped(*args, **kwargs):
            try:
                result = f(*args, **kwargs)
            except expected as err:
                if isexception(replacement):
                    exc = replacement(message) if message else replacement
                    raise exc from err
                return replacement
            return result
        return wrapped
    return wrapper


def isexception(x):
    """True if `x` is a subtype of `Exception`."""
    try:
        truth = issubclass(x, Exception)
    except TypeError:
        return False
    return truth





"""
Create and modify runs in this EPREM project.
"""

import argparse
import typing

from . import etc
from . import interfaces
from . import project
from . import paths


FILEPATH = paths.normalize(__file__)
DIRECTORY = FILEPATH.parent


cli = interfaces.CLI(
    formatter_class=argparse.RawTextHelpFormatter,
    description=etc.doc2help(__doc__),
)


@cli.include
def run(
    config: str,
    target: str=None,
    branches: typing.Union[str, typing.Iterable[str]]=None,
    nproc: int=None,
    force: bool=False,
    verbose: bool=False,
    **environment
) -> None:
    """Set up and execute a new EPREM run in all applicable branches."""
    prj = project.Interface(DIRECTORY)
    prj.run(
        config=config,
        name=target,
        branches=branches,
        nproc=nproc,
        environment=environment,
        errors=(not force),
        silent=(not verbose),
    )
cli.subcommands['run'].add_argument(
    'config',
    help=(
        "the EPREM config file to use"
        ";\nmay be a (possibly relative) path"
        "\nor the name of a file in"
        f"\n{DIRECTORY}/inputs"
    ),
)
cli.subcommands['run'].add_argument(
    '-t',
    '--target',
    help="the name to the new run\n(default: created from date and time)",
)
cli.subcommands['run'].add_argument(
    '-b',
    '--branches',
    help="the affected branches\n(default: all)",
    nargs='*',
    metavar=('A', 'B'),
)
cli.subcommands['run'].add_argument(
    '-n',
    '--nproc',
    help="the number of parallel processes to use\n(default: 1)",
    type=int,
    default=1,
    metavar='N',
)
cli.subcommands['run'].add_argument(
    '-f',
    '--force',
    help="do not raise exceptions when creating a new run",
    action='store_true',
)
cli.subcommands['run'].add_argument(
    '-m',
    '--mpirun',
    help="path to the MPI binary to use\n(default: search $PATH)",
)
cli.subcommands['run'].add_argument(
    '-e',
    '--eprem',
    help="path to the EPREM executable to run\n(default: search $PATH)",
)
cli.subcommands['run'].add_argument(
    '-v',
    '--verbose',
    help="print runtime messages",
    action='store_true',
)


@cli.include
def mv(
    source: str,
    target: str,
    branches: typing.Union[str, typing.Iterable[str]]=None,
    force: bool=False,
    verbose: bool=False,
) -> None:
    """Rename an existing EPREM run in all applicable branches."""
    prj = project.Interface(DIRECTORY)
    prj.mv(
        source,
        target,
        branches=branches,
        errors=(not force),
        silent=(not verbose),
    )
cli.subcommands['mv'].add_argument(
    'source',
    help="the current name of the run(s)",
)
cli.subcommands['mv'].add_argument(
    'target',
    help="the new name of the run(s)",
)
cli.subcommands['mv'].add_argument(
    '-b',
    '--branches',
    help="the affected branches\n(default: all)",
    nargs='*',
    metavar=('A', 'B'),
)
cli.subcommands['mv'].add_argument(
    '-f',
    '--force',
    help="do not raise exceptions when renaming a file",
    action='store_true',
)
cli.subcommands['mv'].add_argument(
    '-v',
    '--verbose',
    help="print runtime messages",
    action='store_true',
)


@cli.include
def rm(
    target: str,
    branches: typing.Union[str, typing.Iterable[str]]=None,
    force: bool=False,
    verbose: bool=False,
) -> None:
    """Remove an existing EPREM run in all applicable branches."""
    prj = project.Interface(DIRECTORY)
    prj.rm(
        target,
        branches=branches,
        errors=(not force),
        silent=(not verbose),
    )
cli.subcommands['rm'].add_argument(
    'target',
    help="the name of the run(s) to remove",
)
cli.subcommands['rm'].add_argument(
    '-b',
    '--branches',
    help="the affected branches\n(default: all)",
    nargs='*',
    metavar=('A', 'B'),
)
cli.subcommands['rm'].add_argument(
    '-f',
    '--force',
    help="do not raise exceptions when removing a file",
    action='store_true',
)
cli.subcommands['rm'].add_argument(
    '-v',
    '--verbose',
    help="print runtime messages",
    action='store_true',
)





from .. import real
from ._context import Context
from ._object import (
    Object,
    ObjectT,
)


class Value(Object[real.ValueType]):
    """A measured object with a single value.

    See Also
    --------
    `~Object`
        An object with real-valued data and an associated unit.
    """

    def __init__(
        self,
        data: real.ValueType,
        context: Context[ObjectT],
        /,
    ) -> None:
        super().__init__(data, context)

    def __int__(self) -> int:
        """Called for int(self)."""
        return int(self.data)

    def __float__(self) -> float:
        """Called for float(self)."""
        return float(self.data)

    def __complex__(self) -> complex:
        """Called for complex(self)."""
        return complex(float(self))





import os
import pathlib
import shutil
import typing

from . import etc


PathLike = typing.Union[str, os.PathLike]


class PathTypeError(Exception):
    """The path is the wrong type."""


class PathOperationError(Exception):
    """This operation is not allowed on the given path(s)."""


def normalize(p: PathLike):
    """Expand and resolve the given path."""
    return pathlib.Path(p).expanduser().resolve()


def localize(p: PathLike, normalized: bool=False):
    """Convert `p` to local path.
    
    This function simply ensures that the return value is an instance of
    ``pathlib.Path`` without parents. It may be useful for ensuring that future
    operations don't write to an arbitrary location on disk.

    Parameters
    ----------
    p : path-like
        A representation of the path to convert.

    normalized : bool, default=False
        If true, expand `~` and fully resolve `p` before extracting its final
        component. By default, this function will create an intermediate path
        from `p` as-is.
    """
    path = normalize(p) if normalized else pathlib.Path(p)
    return pathlib.Path(path.name)


def locate(
    name: str,
    path: pathlib.Path,
    environment: typing.Dict[str, str]
) -> pathlib.Path:
    """Compute an appropriate path to the named element.

    Notes
    -----
    * Intended for use by `~Project`.
    * This function will attempt to create a full path (resolving links as
      necessary) based on `environment` or from `path / name`. If neither exist,
      it will return `name` as-is, thereby allowing calling code to default to
      the searching the system path.
    """
    location = environment.get(name) or path / name
    it = normalize(os.path.realpath(location))
    return it if it.exists() else pathlib.Path(shutil.which(name))


class PathSet(etc.InstanceSet):
    """A metaclass for sets of path-based objects.

    Using this class as the metaclass for a class that manages path-based
    objects (e.g., datasets) will ensure that only one instance of the object
    exists. This may be useful when multiple other objects need to access an
    instance of the class given a common path.

    See https://refactoring.guru/design-patterns/singleton/python/example
    """

    _instances = {}

    def _generate_key(self, *args, **kwargs):
        path = kwargs.get('path')
        if path is None and len(args) >= 1:
            path = args[0]
        if isinstance(path, (str, pathlib.Path)):
            return path
        return super()._generate_key(*args, **kwargs)


class NonExistentPathError(Exception):

    def __init__(self, path: str=None):
        self._path = path

    @property
    def path(self) -> str:
        if self._path is None:
            self._path = "The requested path"
        return self._path

    def __str__(self):
        return f"{self.path} does not exist."


def fullpath(path: PathLike, strict: bool=False):
    """Expand and resolve `path`.

    Parameters
    ----------
    path : string or `os.PathLike`
        The target path.
    strict : bool, default=False
        If true and the computed path does not exist, raise an exception.
        Otherwise, this function will return the new path.

    Returns
    -------
    `pathlib.Path`

    Raises
    ------
    `~NonExistentPathError`: See `strict` keyword.
    """
    new = pathlib.Path(path).expanduser().resolve()
    if strict and not new.exists():
        raise NonExistentPathError(path)
    return new


Parsable = typing.TypeVar('Parsable')
Parsable = typing.Dict[str, typing.Any]


class TextFile:
    """A representation of a text file with pattern-based search."""

    def __init__(self, path: PathLike) -> None:
        """
        Parameters
        ----------
        path : string or `pathlib.Path`
            The path to the target text file. May be relative and contain
            wildcards.
        """
        self._path = path
        self._lines = None

    @property
    def lines(self):
        """The lines of text in this file."""
        if self._lines is None:
            with self.path.open('r') as fp:
                self._lines = fp.readlines()
        return self._lines

    @property
    def path(self):
        """The fully resolved path."""
        return fullpath(self._path)

    KT = typing.TypeVar('KT', bound=typing.Hashable)

    VT = typing.TypeVar('VT')

    Matched = typing.TypeVar('Matched')

    Parsed = typing.TypeVar('Parsed', bound=tuple)

    Parsed = typing.Tuple[KT, VT]

    def extract(
        self,
        match: typing.Callable[[str], Matched],
        parse: typing.Callable[[Matched], Parsed],
    ) -> typing.Dict[KT, VT]:
        """Search each line and parse those that meet given criteria.
        
        Parameters
        ----------
        match : callable
            A callable object that takes a string and returns matches to a
            pattern.
        parse : callable
            A callable object that takes the output of `match`, and returns a
            tuple containing a valid mapping key and corresponding value. It may
            assume that the input is not empty.

        Returns
        -------
        dict
            A dictionary constructed from the tuples output by `parse`.
        """
        matches = [match(line) for line in self.lines]
        parsed = [parse(match) for match in matches if match]
        return dict(parsed)

    def __str__(self) -> str:
        """A simplified representation of this object."""
        return str(self.path)

    def __repr__(self) -> str:
        """An unambiguous representation of this object."""
        return str(self.path)


def file_lines(file: PathLike) -> int:
    """Count the number of lines in a file."""
    with pathlib.Path(file).open('r') as fp:
        nlines = len(fp.readlines())
    return nlines


def strip_inline_comments(string: str, comments: typing.Iterable[str]) -> str:
    """Remove inline comments from a string."""
    for c in comments:
        parts = string.split(c)
        string = parts[0]
    return string.strip()


def search(file: PathLike, *paths: PathLike) -> pathlib.Path:
    """Search `paths` for `file`.
    
    Parameters
    ----------
    file : path-like
        The file to locate

    paths : path-like
        Zero or more paths to search, in the order given. Each member must be an
        object that can represent a path on the current file system. If missing,
        this function will search the current working directory.

    Returns
    -------
    `pathlib.Path` or `None`
        The full path to the file, if found.
    """
    for p in list(paths) + [pathlib.Path.cwd()]:
        path = fullpath(p, strict=True)
        if path.is_dir():
            test = path / str(file)
            if test.exists():
                return test


def find_file_by_template(
    templates: typing.List[typing.Callable],
    name: str,
    directory: PathLike=pathlib.Path.cwd(),
) -> pathlib.Path:
    """Find a valid path that conforms to a given template."""
    d = fullpath(directory, strict=True)
    for template in templates:
        test = d / str(template(name))
        if test.exists():
            return test





"""
Base classes for objects that provide an interface to numeric data.
"""

import collections.abc
import functools
import inspect
import numbers
import typing

from .. import etc
from ..typehelp import Self


ValueType = typing.TypeVar('ValueType', int, float, complex, numbers.Complex)
"""Type variable representing a numeric value.

This type variable is constrained to be one of the following standard types

- `int`
- `float`
- `complex`
- `numbers.Complex`
"""

T = typing.TypeVar('T')

# NOTE: These are a WIP attempt to define a generic type variable that
# represents either a single number or an array-like object of arbitrary depth
# that contains only numbers. So far, it works for single numbers and nested
# lists, but not for `numpy` arrays.
_NestedSequence = typing.Union[
    typing.Sequence[T],
    typing.Sequence["_NestedSequence[T]"],
]
DataType = typing.Union[
    ValueType,
    _NestedSequence[ValueType],
]


class Registry(collections.abc.Mapping, typing.Generic[T]):
    """A collection of decorators for setting numeric-object attributes."""

    def __init__(self, __type: typing.Type[T]) -> None:
        self._type = __type
        self._registered = {'factory': __type}

    @typing.overload
    def factory(*, mutable: typing.Optional[typing.Sequence[str]]=None): ...

    def factory(
        self,
        f: typing.Callable[..., T]=None,
        *,
        mutable: typing.Optional[typing.Sequence[str]]=None,
    ) -> typing.Callable[..., T]:
        """Set the object factory.

        This method is intended for use as a decorator. When called as a
        standard decorator, it will prevent attempts to update attributes when
        the first argument is an existing instance of the class. When called
        with a sequence of attribute names, it will relax the immutability
        restriction on the named attributes.
        """
        keys = mutable or []
        def decorate(c: typing.Callable[..., T]):
            wrapped = _wrap_factory(c, *keys)
            self._registered['factory'] = wrapped
            return wrapped
        if f is None:
            return decorate
        return decorate(f)

    def __len__(self) -> int:
        """Called for len(self)."""
        return len(self._registered)

    def __iter__(self) -> typing.Iterator:
        """Called for iter(self)."""
        return iter(self._registered)

    def __getitem__(self, __k: str):
        """Called for key-based access."""
        try:
            value = self._registered[__k]
        except KeyError as err:
            raise KeyError(
                f"This collection does not include {__k!r}"
            ) from err
        return value


class Registered(typing.Generic[T]):
    """Interface to registered attributes."""

    def __init__(self, registry: Registry[T]) -> None:
        self.registry = registry

    @property
    def factory(self) -> typing.Callable[..., T]:
        """The registered object factory, if any."""
        return self._getattr('factory')

    def _getattr(self, name: str):
        """Internal attribute look-up method."""
        try:
            this = self.registry[name]
        except KeyError as err:
            raise AttributeError(name) from err
        return this


class Mixin:
    """Class methods and properties for numeric objects."""

    register: typing.ClassVar[Registry[Self]]=None

    @classmethod
    def spawn(cls, x, **metadata):
        """Create a new instance of this type."""
        return cls._spawn(x, **metadata)

    @classmethod
    def _spawn(cls, *args, **kwargs):
        """Create a new instance of this type."""
        if cls.has('factory'):
            return cls.registered.factory(*args, **kwargs)
        raise NotImplementedError(
            f"{cls} does not have an associated instance factory"
        ) from None

    @classmethod
    def has(cls, name: str):
        """True if this class has registered the named attribute."""
        if cls.register is None:
            return False
        return cls.register.get(name) is not None

    @etc.classproperty
    def registered(cls):
        """The collection of registered attributes."""
        return Registered(cls.register)


class Quantity(Mixin, typing.Generic[T]):
    """Base class for generic interfaces."""

    def __init_subclass__(cls, **kwargs) -> None:
        """Defined to reset the attribute registry on each subclass."""
        super().__init_subclass__(**kwargs)
        cls.register = Registry(cls)

    def __init__(self, __object: T) -> None:
        self._object = __object

    def __repr__(self) -> str:
        return f"{self.__class__.__qualname__}[{type(self._object)}]"


@etc.autostr
class Interface(typing.Protocol[T]):
    """Abstract protocol class for interfaces to numeric data.

    This class defines an internal `_data` attribute and a public `data`
    property, and provides basic `__str__` and `__repr__` methods.
    """

    _data: T

    @property
    def data(self) -> T:
        """This object's underlying data."""
        return self._data


InterfaceT = typing.TypeVar('InterfaceT', bound=Interface)
"""Type variable bounded by `~numeric.Interface`."""


class Object(Interface[T], Mixin):
    """A concrete interface to numeric data.

    This class inherits the internal `_data` attribute and the public `data`
    property from `~Interface`. It extends that class by providing a concrete
    initialization method, as well as basic string-representation methods.
    """

    def __init_subclass__(cls, **kwargs) -> None:
        """Defined to reset the attribute registry on each subclass."""
        super().__init_subclass__(**kwargs)
        cls.register = Registry(cls)

    def __init__(self, data: T) -> None:
        """Initialize this object with `data`."""
        self._data = data

    def __str__(self) -> str:
        """A simplified representation of this object."""
        return str(self.data)


ObjectT = typing.TypeVar('ObjectT', bound=Object)
"""Type variable bounded by `~numeric.Object`."""


@typing.runtime_checkable
class Type(typing.Protocol[T]):
    """Structural protocol class for generic data-centric objects.

    Objects that define a `_data` attribute and a `data` property will pass
    instance checks against this protocol.
    """

    _data: T

    @property
    def data(self) -> T: ...


def _wrap_factory(f: typing.Callable[..., ObjectT], *keys: str):
    """Internal logic for checking numeric-object factories."""
    rettype = inspect.signature(f).return_annotation
    newtype = Object if rettype is inspect.Signature.empty else rettype
    @functools.wraps(f)
    def wrapped(x, /, **kwargs):
        rtstr = newtype.__qualname__
        xtstr = type(x).__qualname__
        rest = (
            f"when creating an instance of {rtstr}"
            f" from an existing instance of {xtstr}"
        )
        errmsg = "Cannot change {} " + rest
        if isinstance(x, Object):
            for k, v in kwargs.items():
                if k not in keys and v is not None:
                    raise ValueError(errmsg.format(repr(k))) from None
        return f(x, **kwargs)
    return wrapped





"""
Abstract protocols and base classes for general nunmeric quantities.
"""

from . import abc
from . import mixins
from ._protocols import (
    Additive,
    Algebraic,
    Array,
    Comparable,
    Complex,
    HasData,
    HasUnit,
    Multiplicative,
    Measurable,
    Measurement,
    Ordered,
    Real,
    Scalar,
    Sequence,
    Variable,
)
from ._types import (
    ArrayType,
    ComplexType,
    MeasurableType,
    MeasuredType,
    MeasurementType,
    MetricType,
    RealType,
    ScalarType,
    SequenceType,
)


__all__ = [
    # modules
    'abc',
    'mixins',
    # protocol classes
    'Additive',
    'Algebraic',
    'Array',
    'Comparable',
    'Complex',
    'HasData',
    'HasUnit',
    'Multiplicative',
    'Measurable',
    'Measurement',
    'Ordered',
    'Real',
    'Scalar',
    'Sequence',
    'Variable',
    # type variables
    'ArrayType',
    'ComplexType',
    'MeasurableType',
    'MeasuredType',
    'MeasurementType',
    'MetricType',
    'RealType',
    'ScalarType',
    'SequenceType',
]


def isordered(this, /):
    """True if `this` is an ordered object."""
    return isinstance(this, _protocols.Ordered)


def iscomparable(this, /):
    """True if `this` is a comparable object."""
    return isinstance(this, _protocols.Comparable)


def isadditive(this, /):
    """True if `this` is an additive object."""
    return isinstance(this, _protocols.Additive)


def ismultiplicative(this, /):
    """True if `this` is a multiplicative object."""
    return isinstance(this, _protocols.Multiplicative)


def isalgebraic(this, /):
    """True if `this` is an algebraic object."""
    return isinstance(this, _protocols.Algebraic)


def iscomplex(this, /):
    """True if `this` is a complex-valued object."""
    return isinstance(this, _protocols.Complex)


def isreal(this, /):
    """True if `this` is a real-valued object."""
    return isinstance(this, _protocols.Real)


def issequence(this, /):
    """True if `this` is an sequence-like object."""
    return isinstance(this, _protocols.Sequence)


def isarray(this, /):
    """True if `this` is an array-like object."""
    return isinstance(this, _protocols.Array)


def isscalar(this, /):
    """True if `this` is a scalar object."""
    return isinstance(this, _protocols.Scalar)


def isvariable(this, /):
    """True if `this` is a variable object."""
    return isinstance(this, _protocols.Variable)


def hasdata(this, /):
    """True if `this` has a `data` property."""
    return isinstance(this, _protocols.HasData)


def hasunit(this, /):
    """True if `this` has a `unit` property."""
    return isinstance(this, _protocols.HasUnit)


def ismeasured(this, /):
    """True if `this` is a measured object."""
    return isinstance(this, _protocols.Measured)


def ismeasurement(this, /):
    """True if `this` is a measurement-like object."""
    return isinstance(this, _protocols.Measurement)


def ismeasurable(this, /):
    """True if `this` is an explicitly measurable object."""
    return isinstance(this, _protocols.Measurable)






import abc
import collections.abc
import contextlib
import numbers
import typing

from .. import aliased
from .. import etc
from .. import symbolic
from . import _exceptions
from . import _reference


Instance = typing.TypeVar('Instance', bound='_Property')


@etc.autostr
class _Property(collections.abc.Mapping):
    """All definitions of a single metric property."""

    _instances = {}
    _supported = (
        'dimensions',
        'units',
    )

    key: str=None
    _cache: dict=None

    def __new__(
        cls: typing.Type[Instance],
        arg: typing.Union[str, Instance],
    ) -> Instance:
        """Create a new instance or return an existing one.
        
        Parameters
        ----------
        arg : string or instance
            A string representing the metric property to create, or an existing
            instance of this class.
        """
        if isinstance(arg, cls):
            return arg
        key = str(arg)
        if key not in cls._supported:
            raise ValueError(f"Unsupported property: {key}") from None
        if available := cls._instances.get(key):
            return available
        self = super().__new__(cls)
        self.key = key
        self._cache = {}
        cls._instances[key] = self
        return self

    def system(self, system: str):
        """Get all definitions of this property for `system`."""
        return {k: v[system] for k, v in self.items()}

    LEN = len(_reference._QUANTITIES) # No need to compute every time.

    def __len__(self) -> int:
        """The number of defined quantities. Called for len(self)."""
        return self.LEN

    def __iter__(self) -> typing.Iterator[str]:
        """Iterate over names of defined quantities. Called for iter(self)."""
        return iter(_reference._QUANTITIES)

    def __getitem__(self, quantity: str) -> typing.Dict[str, str]:
        """Create or retrieve a named property."""
        if quantity in self._cache:
            return self._cache[quantity]
        if new := self._get_property(quantity):
            self._cache[quantity] = new
            return new
        raise KeyError(f"Unknown quantity {quantity}") from None

    def _get_property(self, quantity: str) -> typing.Dict[str, str]:
        """Get this property for a defined quantity.
        
        This method will search for `quantity` in the module-level collection of
        defined quantities. If it doesn't find an entry, it will attempt to
        parse `quantity` into known quantities. If it finds a `dict` entry, it
        will attempt to extract the values corresponding to this property's
        `key` attribute (i.e., 'units' or 'dimensions'). If it finds a `str`
        entry, it will attempt to create the equivalent `dict` by symbolically
        evaluating the terms in the entry.
        """
        if quantity not in _reference._QUANTITIES:
            return self._parse(quantity)
        q = _reference._QUANTITIES[quantity]
        if isinstance(q, dict):
            return q.get(self.key, {})
        if not isinstance(q, str):
            raise TypeError(f"Expected {quantity} to be a string") from None
        return self._parse(q)

    def _parse(self, string: str):
        """Parse a string representing a compound quantity."""
        for k in _reference._QUANTITIES:
            string = string.replace(k, k.replace(' ', '_'))
        parts = [
            self._expand(term)
            for term in symbolic.expression(string)
        ]
        keys = {key for part in parts for key in part.keys()}
        merged = {key: [] for key in keys}
        for part in parts:
            for key, value in part.items():
                merged[key].append(value)
        return {
            k: str(symbolic.expression(v))
            for k, v in merged.items()
        }

    _operand = symbolic.OperandFactory()
    def _expand(self, term: symbolic.Term):
        """Create a `dict` of operands from this term."""
        return {
            k: self._operand.create(v, term.exponent)
            for k, v in self[term.base.replace('_', ' ')].items()
        }

    def __str__(self) -> str:
        """A simplified representation of this object."""
        return self.key


# NOTE: Defining mappings from unit or dimension to quantity is a bad idea
# because some quantities have the same unit or dimension in a given system.
# This makes the mapping ill-defined. Python dictionaries simply use the latest
# entry for a repeated key, which means some quantities would overwrite others.
# The following objects rely on mappings from quantity to unit or dimension,
# which are always well defined.


DIMENSIONS = _Property('dimensions')
"""All defined metric dimensions.

This mapping is keyed by physical quantity followed by metric system.
"""


UNITS = _Property('units')
"""All defined metric units.

This mapping is keyed by physical quantity followed by metric system.
"""


CANONICAL = {
    k: {
        system: _Property(k).system(system) for system in ('mks', 'cgs')
    } for k in ('dimensions', 'units')
}
"""Canonical metric properties in each known metric system.

This mapping is keyed by {'dimensions', 'units'}, followed by metric system, and
finally by physical quantity.
"""


MKS = {k: CANONICAL[k]['mks'].copy() for k in ('dimensions', 'units')}
"""All defined units in the MKS system, keyed by physical quantity."""


CGS = {k: CANONICAL[k]['cgs'].copy() for k in ('dimensions', 'units')}
"""All defined units in the CGS system, keyed by physical quantity."""


_EQUIVALENT = {
    p: {
        k: {
            s: symbolic.expression(x)
            for s, x in v.items()
        } for k, v in dict(_Property(p)).items()
    } for p in ('dimensions', 'units')
}

EQUIVALENT = aliased.Mapping(_EQUIVALENT)


class Prefix(typing.NamedTuple):
    """Metadata for a metric order-of-magnitude prefix."""

    symbol: str
    name: str
    factor: float


class BaseUnit(typing.NamedTuple):
    """Metadata for a named unit without metric prefix."""

    symbol: str
    name: str
    quantity: str


Instance = typing.TypeVar('Instance', bound='NamedUnit')


class _NamedUnitMeta(abc.ABCMeta):
    """Internal metaclass for `~metric.NamedUnit`.
    
    This class exists to create singleton instances of `~metric.NamedUnit`
    without needing to overload `__new__` on that class or its base class(es).
    """

    _instances = aliased.MutableMapping()
    _attributes = {}

    def __call__(
        cls,
        arg: typing.Union[Instance, symbolic.Expressable],
    ) -> Instance:
        """Create a new instance or return an existing one."""
        if isinstance(arg, cls):
            # If the argument is already an instance, return it.
            return arg
        string = str(arg)
        if available := cls._instances.get(string):
            # If the argument maps to an existing unit, return that unit.
            return available
        # First time through: identify the base unit and prefix.
        magnitude, reference = identify(string)
        # Store the information to initialize a new instance.
        name = f"{magnitude.name}{reference.name}"
        symbol = f"{magnitude.symbol}{reference.symbol}"
        cls._attributes[string] = {
            'prefix': magnitude,
            'base': reference,
            'name': name,
            'symbol': symbol,
            'scale': magnitude.factor,
            'quantity': reference.quantity,
        }
        # Create the new instance. This will ultimately pass control to
        # `NamedUnit.__init__`, which will initialize the newly instantiated
        # instance with the stored attributes corresponding to `str(arg)`.
        instance = super().__call__(arg)
        cls._instances[(name, symbol)] = instance
        return instance


def identify(string: str):
    """Determine the magnitude and reference unit, if possible.
    
    Parameters
    ----------
    string : str
        A string representing a metric unit.

    Returns
    -------
    tuple
        A 2-tuple in which the first element is a `~metric.Prefix` representing
        the order-of-magnitude of the given unit and the second element is a
        `~metric.BaseUnit` representing the unscaled (i.e., order-unity) metric
        unit.

    Examples
    --------
    >>> mag, ref = identify('km')
    >>> mag
    Prefix(symbol='k', name='kilo', factor=1000.0)
    >>> ref
    BaseUnit(symbol='m', name='meter', quantity='length')
    """
    try:
        unit = _reference.NAMED_UNITS[string]
    except KeyError as err:
        raise _exceptions.UnitParsingError(string) from err
    magnitude = Prefix(**unit['prefix'])
    reference = BaseUnit(**unit['base'])
    return magnitude, reference


@etc.autostr
class Reduction:
    """The components of a reduced unit expression."""

    def __init__(
        self,
        terms: symbolic.Expressable,
        scale: float=1.0,
        system: str=None,
    ) -> None:
        self._expression = scale * symbolic.expression(terms)
        self.system = system
        self.scale = scale
        self._units = None

    @property
    def units(self) -> typing.List[symbolic.Term]:
        """The unit terms in this reduction."""
        if self._units is None:
            self._units = [
                unit for unit in self._expression.terms
                if unit.base != '1'
            ]
        return self._units

    def __mul__(self, other):
        """Called for self * other."""
        if not isinstance(other, numbers.Real):
            return NotImplemented
        scale = self.scale * other
        terms = list(scale * self._expression)
        return type(self)(terms, scale=scale, system=self.system)

    __rmul__ = __mul__
    """Called for other * self."""

    def __truediv__(self, other):
        """Called for self / other."""
        if not isinstance(other, numbers.Real):
            return NotImplemented
        scale = self.scale / other
        terms = list(scale * self._expression)
        return type(self)(terms, scale=scale, system=self.system)

    def __pow__(self, other):
        """Called for self ** other."""
        if not isinstance(other, numbers.Real):
            return NotImplemented
        scale = self.scale ** other
        terms = list(scale * self._expression**other)
        return type(self)(terms, scale=scale, system=self.system)

    def __str__(self) -> str:
        return f"{self._expression} [{self.system!r}]"


@etc.autostr
class NamedUnit(metaclass=_NamedUnitMeta):
    """A single named unit and corresponding metadata."""

    @typing.overload
    def __init__(
        self: Instance,
        unit: str,
    ) -> Instance:
        """Create a new instance or return an existing one.
        
        Parameters
        ----------
        unit : string
            A string representing the metric unit to create.
        """

    @typing.overload
    def __init__(
        self: Instance,
        instance: Instance,
    ) -> Instance:
        """Create a new instance or return an existing one.
        
        Parameters
        ----------
        instance : `~metric.NamedUnit`
            An existing instance of this class.
        """

    def __init__(self, arg) -> None:
        self._parsed = self.__class__._attributes[str(arg)]
        self._prefix = None
        self._base = None
        self._name = None
        self._symbol = None
        self._scale = None
        self._quantity = None
        self._systems = None
        self._dimensions = None
        self._decomposed = None
        self._norm = None
        self._reductions = dict.fromkeys(_reference.SYSTEMS)

    @property
    def norm(self):
        """The equivalent unit, represented in base units of `system`.
        
        Notes
        -----
        This property returns a copy of the original `dict` of normalized units
        in order to prevent modifying singleton instances.
        """
        if self._norm is None:
            self._norm = {
                system: type(self)(UNITS[self.quantity][system])
                for system in _reference.SYSTEMS
            }
        return self._norm.copy()

    @property
    def prefix(self) -> Prefix:
        """The order of magnitide of this unit's metric prefix."""
        if self._prefix is None:
            self._prefix = self._parsed["prefix"]
        return self._prefix

    @property
    def base(self) -> BaseUnit:
        """The reference unit without metric prefix."""
        if self._base is None:
            self._base = self._parsed["base"]
        return self._base

    @property
    def name(self) -> str:
        """The full name of this unit."""
        if self._name is None:
            self._name = self._parsed["name"]
        return self._name

    @property
    def symbol(self) -> str:
        """The abbreviated symbol for this unit."""
        if self._symbol is None:
            self._symbol = self._parsed["symbol"]
        return self._symbol

    @property
    def scale(self) -> float:
        """The metric scale factor of this unit."""
        if self._scale is None:
            self._scale = self._parsed["scale"]
        return self._scale

    @property
    def quantity(self) -> str:
        """The physical quantity of this unit."""
        if self._quantity is None:
            self._quantity = self._parsed["quantity"]
        return self._quantity

    @property
    def systems(self):
        """The metric systems that use this unit.
        
        This property uses the criteria described in
        `~metric.NamedUnit.is_allowed_in` to build the collection of metric
        systems, most notably that named units not defined in any metric system
        are allowed in all metric systems.
        """
        if self._systems is None:
            modes = {
                k: []
                for k in {'allowed', 'defined', 'fundamental'}
            }
            for system in _reference.SYSTEMS:
                if self.is_fundamental_in(system):
                    modes['fundamental'].append(system)
                if self.is_defined_in(system):
                    modes['defined'].append(system)
                if self.is_allowed_in(system):
                    modes['allowed'].append(system)
            self._systems = {k: tuple(v) for k, v in modes.items()}
        return self._systems.copy()

    @property
    def decomposed(self):
        """The representation of this unit in base units, if possible."""
        if self._decomposed is None:
            with contextlib.suppress(StopIteration):
                system = next(
                    system for system in _reference.SYSTEMS
                    if self.is_fundamental_in(system)
                )
                self._decomposed = self._decompose(system)
        return self._decomposed

    def _decompose(self, system: typing.Literal['mks', 'cgs']):
        """Internal logic for `NamedUnit.decomposed`."""
        if not self.is_defined_in(system):
            # If this unit is not defined in this metric system, we can't
            # decompose it.
            return
        dimension = self.dimensions[system]
        expression = symbolic.expression(dimension)
        if len(dimension) == 1:
            # If this unit's dimension is irreducible, there's no point in going
            # through all the decomposition logic.
            return [symbolic.term(self.symbol)]
        quantities = [
            _reference._BASE_QUANTITIES.find(term.base, unique=True)['name']
            for term in expression
        ]
        units = [
            _reference._QUANTITIES[quantity]['units'][system]
            for quantity in quantities
        ]
        return [
            symbolic.term(base=unit, exponent=term.exponent)
            for unit, term in zip(units, expression)
        ]

    def reduce(self, system: str=None) -> typing.Optional[Reduction]:
        """Convert this unit to base units of `system`, if possible."""
        s = self._resolve_system(system)
        if self._reductions[s]:
            return self._reductions[s]
        result = self._reduce(s)
        self._reductions[s] = result
        return result

    def _resolve_system(self, system: typing.Optional[str]):
        """Determine the appropriate metric system to use, if possible."""
        if isinstance(system, str) and system.lower() in _reference.SYSTEMS:
            # trivial case
            return system.lower()
        systems = [s for s in _reference.SYSTEMS if self.is_fundamental_in(s)]
        if len(systems) == 1:
            # use canonical system if possible
            return systems[0]
        if self.dimensions['mks'] == self.dimensions['cgs']:
            # system-independent: use mks by default
            return 'mks'
        if self.dimensions['mks'] is None:
            # only defined in cgs
            return 'cgs'
        if self.dimensions['cgs'] is None:
            # only defined in mks
            return 'mks'
        # system-dependent but we don't know the system
        raise _exceptions.SystemAmbiguityError(str(self))

    def _reduce(self, system: typing.Literal['mks', 'cgs']):
        """Internal logic for `~NamedUnit.reduce`."""
        if not self.is_defined_in(system):
            # If this unit is not defined in this metric system, we can't reduce
            # it.
            return
        dimension = self.dimensions[system]
        expression = symbolic.expression(dimension)
        if len(expression) == 1:
            # If this unit's dimension is irreducible, there's no point in going
            # through all the reduction logic.
            canonical = CANONICAL['units'][system][self.quantity]
            if self.symbol == canonical:
                # If this is the canonical unit for its quantity in `system`,
                # return it with a scale of unity.
                return Reduction(
                    [symbolic.term(self.symbol)],
                    system=system,
                )
            # If not, return the canonical unit with the appropriate scale
            # factor.
            return Reduction(
                [symbolic.term(canonical)],
                scale=(canonical << self),
                system=system,
            )
        quantities = [
            _reference._BASE_QUANTITIES.find(term.base, unique=True)['name']
            for term in expression
        ]
        units = [
            _reference._QUANTITIES[quantity]['units'][system]
            for quantity in quantities
        ]
        terms = [
            symbolic.term(base=unit, exponent=term.exponent)
            for unit, term in zip(units, expression)
        ]
        return Reduction(terms, scale=self.scale, system=system)

    @property
    def dimensions(self) -> typing.Dict[str, typing.Optional[str]]:
        """The physical dimension of this unit in each metric system.
        
        Notes
        -----
        This property returns a copy of an internal `dict` in order to prevent
        accidentally changing the instance dimensions through an otherwise valid
        `dict` operation. Such changes are irreversible since each
        `~metric.NamedUnit` instance is a singleton.
        """
        if self._dimensions is None:
            systems = {system for system in self.systems['defined']}
            self._dimensions = {
                k: symbolic.expression(v) or None
                for k, v in self._get_dimensions(systems).items()
            }
        return self._dimensions.copy()

    def _get_dimensions(self, systems: typing.Set[str]):
        """Helper for computing dimensions of this named unit.
        
        Notes
        -----
        This method requires the full set of applicable metric systems (rather
        than one system at a time) because it will return all available
        dimensions if either 1) there are no applicable systems, or 2) the set
        of applicable systems is equal to the set of all known systems.
        """
        dimensions = DIMENSIONS[self.quantity]
        if not systems or (systems == _reference.SYSTEMS):
            return dimensions.copy()
        base = dict.fromkeys(_reference.SYSTEMS)
        if len(systems) == 1:
            system = systems.pop()
            base[system] = dimensions[system]
            return base
        raise _exceptions.SystemAmbiguityError

    def is_allowed_in(self, system: typing.Literal['mks', 'cgs']):
        """True if this named unit inter-operates with units in `system`.
        
        A named unit is allowed in some or all metric systems, but never none.
        The reason for this is that a named unit that is not defined in any
        metric system is effectively independent of all metric systems, so
        attempting to restrict its use to a subset of metric systems is
        fruitless.

        See Also
        --------
        `~is_defined_in`
            True if the given metric system formally contains this named unit.

        `~is_fundamental_in`
            True if this named unit is the fundamental unit for its dimension in
            the given metric system.
        """
        systems = {
            s for s in _reference.SYSTEMS if self.is_defined_in(s)
        } or _reference.SYSTEMS
        return system in systems

    def is_defined_in(self, system: typing.Literal['mks', 'cgs']):
        """True if this named unit is defined in `system`."""
        if self.is_fundamental_in(system):
            return True
        canonical = CANONICAL['units'][system][self.quantity]
        with contextlib.suppress(_exceptions.UnitParsingError):
            reference = type(self)(canonical)
            if self.base == reference.base:
                return True
        return False

    def is_fundamental_in(self, system: typing.Literal['mks', 'cgs']):
        """True if this named unit is the canonical unit in `system`."""
        canonical = CANONICAL['units'][system][self.quantity]
        keys = (self.symbol, self.name)
        for key in keys:
            if key == canonical:
                return True
        return False

    def __eq__(self, other) -> bool:
        """True if two representations have equal magnitude and base unit."""
        that = type(self)(other)
        same_magnitude = (self.prefix == that.prefix)
        same_reference = (self.base == that.base)
        return same_magnitude and same_reference

    def __hash__(self) -> int:
        """Called for hash(self). Supports use as mapping key."""
        return hash((self.base, self.prefix))

    def __rshift__(self, other):
        """Compute the conversion factor from this unit to `other`.

        This operation computes the numerical factor necessary to convert a
        quantity in terms of this unit to a quantity in terms of `other`,
        provided both units have the same base unit.

        Examples
        --------
        The result of this operation is equal to the result of `~metric.ratio`:

        >>> metric.ratio('cm', 'm')
        0.01
        >>> metric.NamedUnit('cm') >> 'm'
        0.01
        """
        if isinstance(other, (NamedUnit, str)):
            return ratio(self, other)
        return NotImplemented

    def __rrshift__(self, other):
        """Reflected version of `__rshift__`."""
        if isinstance(other, str):
            return ratio(other, self)
        return NotImplemented

    def __lshift__(self, other):
        """Compute the conversion factor from `other` to this unit.
        
        This operation computes the numerical factor necessary to convert a
        quantity in terms of `other` to a quantity in terms of this unit,
        provided both units have the same base unit.

        Examples
        --------
        The result of this operation is equal to the inverse of the result of
        `~metric.ratio`:

        >>> metric.ratio('cm', 'm')
        0.01
        >>> metric.NamedUnit('cm') << 'm'
        100.0
        """
        if isinstance(other, (NamedUnit, str)):
            return ratio(other, self)
        return NotImplemented

    def __rlshift__(self, other):
        """Reflected version of `__lshift__`."""
        if isinstance(other, str):
            return ratio(self, other)
        return NotImplemented

    def __str__(self) -> str:
        """A printable representation of this unit."""
        return f"'{self.name} | {self.symbol}'"


def ratio(
    this: typing.Union[str, NamedUnit],
    that: typing.Union[str, NamedUnit],
) -> float:
    """Compute the magnitude of `this` relative to `that`.

    Parameters
    ----------
    this : string or `~metric.NamedUnit`
        The reference unit.

    that : string or `~metric.NamedUnit`
        The unit to compare to `this`. Must have the same base unit as `this`.

    Examples
    --------
    The following are all equivalent to the fact that a meter represents 100
    centimeters:

        >>> ratio('meter', 'centimeter')
        100.0
        >>> ratio('centimeter', 'meter')
        0.01
        >>> ratio('m', 'cm')
        100.0
        >>> ratio('cm', 'm')
        0.01

    Attempting this operation between two units with different base units will
    raise an exception,

        >>> ratio('m', 's')
        <raises ValueError>

    even if they represent the same quantity.

        >>> ratio('m', 'au')
        <raises ValueError>

    Therefore, this function is not intended for unit conversion.
    """
    u = [NamedUnit(i) if isinstance(i, str) else i for i in (this, that)]
    if not all(isinstance(i, NamedUnit) for i in u):
        raise TypeError(
            f"Each unit must be an instance of {str!r} or {NamedUnit!r}"
        ) from None
    u0, u1 = u
    if u1 == u0:
        return 1.0
    if u1.base != u0.base:
        units = ' to '.join(f"{i.symbol!r} ({i.name})" for i in u)
        raise ValueError(f"Can't compare {units}") from None
    return u0.scale / u1.scale


def get_quantity(unit: symbolic.Expression):
    """Compute the physical quantity of the given unit."""
    terms = [
        symbolic.term(
            NamedUnit(term.base).quantity.replace(' ', '_')
        ) ** term.exponent
        for term in unit
    ]
    return symbolic.expression(terms)





import typing

from ..metric import (
    Conversion,
)
from ._object import ObjectT


T = typing.TypeVar('T')


class Converter:
    """A callable interface to unit-conversion implementations."""

    def __init__(
        self,
        convert: typing.Callable[[Conversion, ObjectT], ObjectT],
    ) -> None:
        self._convert = convert

    def __call__(self, c: Conversion, x: ObjectT):
        """Apply conversion `c` to object `x`."""
        if c.old == c.new:
            return x
        return self._convert(c, x)





"""
Support for metric dimensions.
"""

import typing

from .. import etc
from .. import symbolic
from . import _defined
from . import _reference


Instance = typing.TypeVar('Instance', bound='Dimension')


class Dimension(symbolic.Expression):
    """An symbolic expression representing a physical dimension."""

    def __init__(self, arg: symbolic.Expressable) -> None:
        super().__init__(arg)
        self._quantities = {}

    def quantities(self, system: str) -> typing.Set[str]:
        """The quantities with this dimension in `system`."""
        if system in self._quantities:
            return self._quantities[system]
        canonical = _defined.CANONICAL['dimensions'][system]
        found = {k for k, v in canonical.items() if v == self}
        self._quantities[system] = found
        return found


def dimension_factory(arg: symbolic.Expressable):
    """Factory function for metric dimensions."""
    return Dimension(symbolic.expression(arg))


@etc.autostr
class Dimensions(typing.Mapping):
    """A collection of symbolic expressions of metric dimensions."""

    def __init__(
        self,
        common: symbolic.Expressable=None,
        **systems: symbolic.Expressable
    ) -> None:
        """Initialize from expressable quantities.
        
        Parameters
        ----------
        common : string or iterable or `~symbolic.Expression`
            The dimension to associate with all metric systems.

        **systems
            Zero or more key-value pairs in which the key is the name of a known
            metric system and the value is an object that can instantiate the
            `~symbolic.Expression` class. If present, each value will override
            `common` for the corresponding metric system.
        """
        self._objects = self._init_from(common, **systems)

    def _init_from(
        self,
        common,
        **systems
    ) -> typing.Dict[str, typing.Optional[Dimension]]:
        """Create dimension objects from arguments."""
        created = dict.fromkeys(_reference.SYSTEMS)
        default = Dimension(common) if common else None
        updates = {
            system: Dimension(expressable) if expressable else default
            for system, expressable in systems.items()
        }
        created.update(updates)
        if any(created.values()):
            return created
        raise TypeError(
            f"Can't instantiate {self.__class__!r}"
            f" from {common!r} and {systems!r}"
        ) from None

    def __len__(self) -> int:
        return len(self._objects)

    def __iter__(self) -> typing.Iterator:
        return iter(self._objects)

    def __getitem__(self, __k):
        key = str(__k).lower()
        if key in self._objects:
            return self._objects[key]
        raise KeyError(f"No dimension for {__k!r}") from None

    def __str__(self) -> str:
        return ', '.join(
            f"{k!r}: {str(v) if v else v!r}"
            for k, v in self._objects.items()
        )





import numpy

from ..metric import Conversion
from ._object import ObjectT


def arraylike(c: Conversion, x: ObjectT):
    """Unit-conversion implementation for array-like objects."""
    return apply(_scale_array, c, x)


def singular(c: Conversion, x: ObjectT):
    """Unit-conversion implementation for singular objects."""
    return apply(_scale_data, c, x)


def _scale_array(c: Conversion, x: ObjectT):
    """Convert `x.data` to an array before applying a conversion factor."""
    return float(c) * numpy.array(x)


def _scale_data(c: Conversion, x: ObjectT):
    """Directly apply a conversion factor to `x.data`."""
    return float(c) * x.data


def apply(f, c: Conversion, x: ObjectT):
    """Apply the given unit conversion to the given object."""
    if c.old == c.new:
        return x
    return x.spawn(f(c, x), unit=c.new)




import argparse
import functools
import types
import typing

from . import etc


class CLI(typing.Mapping):
    """A custom command-line parser for EPREM simulations."""

    def __init__(self, *args, **kwargs):
        """Create a new instance."""
        self._args = args
        self._kwargs = kwargs
        self._parser = argparse.ArgumentParser(*self._args, **self._kwargs)
        self._subparsers = None
        self._subcommands = None

    def __len__(self) -> int:
        """Called for len(self)."""
        return len(self.subcommands)

    def __iter__(self) -> typing.Iterator[str]:
        """Called for iter(self)."""
        return iter(self.subcommands)

    def __getitem__(self, __k: str) -> argparse.ArgumentParser:
        """Access subcommands by key."""
        if __k in self.subcommands:
            return self.subcommands[__k]
        raise KeyError(f"No subcommand for {__k!r}")

    def include(self, _func=None, **meta):
        """Register a subcommand."""
        def cli_action(func: types.FunctionType):
            """Decorate `func` as a command-line action."""
            key = func.__name__
            subparser = self.subparsers.add_parser(
                key,
                help=etc.doc2help(func),
                description=etc.doc2help(func),
                formatter_class=argparse.RawTextHelpFormatter,
                **meta
            )
            subparser.set_defaults(func=func)
            self.subcommands[key] = subparser
            @functools.wraps(func)
            def wrapper(*args, **kwargs):
                return func(*args, **kwargs)
            return wrapper
        if _func is None:
            return cli_action
        return cli_action(_func)

    def run(self):
        """Execute operations based on command-like arguments."""
        parsed = vars(self.parser.parse_args())
        func = parsed.pop('func')
        func(**parsed)

    @property
    def parser(self):
        """The main argument parser."""
        return self._parser

    @property
    def subparsers(self):
        """The parser for each subcommand."""
        if self._subparsers is None:
            self._subparsers = self.parser.add_subparsers(
                title="supported sub-commands",
            )
        return self._subparsers

    @property
    def subcommands(self) -> typing.Dict[str, argparse.ArgumentParser]:
        """The registered subcommands."""
        if self._subcommands is None:
            self._subcommands = {}
        return self._subcommands





"""
EPREM observer interfaces.
"""

import collections.abc
import contextlib
import pathlib
import typing

from . import aliased
from . import datafile
from . import etc
from . import metric
from . import observable
from . import parameter
from . import paths
from . import physical
from . import reference


__all__ = [
    "stream",
    "Stream",
    "point",
    "Point",
    "observer",
    "Observer",
]


BASETYPES = parameter.reference.BASETYPES_H
_CC = aliased.Mapping(parameter.reference.CONFIGURATION_C)
PARAMETERS = [str(group) for group in _CC.keys(aliased=True)]
OBSERVABLES = [str(group) for group in reference.QUANTITIES.keys(aliased=True)]


@etc.autostr
class Observer(collections.abc.Mapping):
    """A generic EPREM observer.

    An instance of this class provides an interface to observable array-like
    quantities via a mapping-style interface.
    """

    def __init__(
        self,
        dataview: datafile.View,
        observables: observable.Quantities,
    ) -> None:
        self._dataview = dataview
        self._observables = observables
        self._source = None
        self._system = None
        self._sizes = None
        self._axes = None
        self._times = None
        self._species = None
        self._energies = None
        self._mus = None

    def __hash__(self):
        """Called for hash(self)."""
        return hash(self.source)

    def __str__(self) -> str:
        """Called for str(self)."""
        return str(self.source)

    def __len__(self) -> int:
        """Called for len(self)."""
        return len(self._observables)

    def __iter__(self):
        """Called for iter(self)."""
        return iter(tuple(self._observables))

    def __getitem__(self, key: str, /):
        """Retrieve the named quantity, if possible."""
        if key in self._observables:
            return self._observables[key]
        raise KeyError(
            f"No observable quantity for {key!r}"
        ) from None

    @property
    def times(self) -> physical.Coordinates:
        """This observer's time coordinates."""
        if self._times is None:
            self._times = self._get_axis('time')
        return self._times

    @property
    def species(self) -> physical.Symbols:
        """This observer's species symbols."""
        if self._species is None:
            self._species = self._get_axis('species')
        return self._species

    @property
    def energies(self) -> physical.Coordinates:
        """This observer's energy coordinates."""
        if self._energies is None:
            self._energies = self._get_axis('energy')
        return self._energies

    @property
    def mus(self) -> physical.Coordinates:
        """This observer's mu coordinates."""
        if self._mus is None:
            self._mus = self._get_axis('mu')
        return self._mus

    def _get_axis(self, name: str):
        """Internal helper for axis properties."""
        if self._axes is None:
            self._axes = datafile.axes(self.source, self.system)
        return self._axes[name]

    @property
    def source(self):
        """The directory containing this observer's dataset."""
        if self._source is None:
            self._source = self.dataview.source
        return self._source

    @property
    def system(self):
        """This observer's metric system."""
        if self._system is None:
            self._system = self._observables.system
        return self._system

    @property
    def dataview(self):
        """An interface to this observer's raw dataset."""
        return self._dataview


ObserverType = typing.TypeVar('ObserverType', bound=Observer)


@etc.autostr
class Stream(Observer):
    """An EPREM stream observer."""

    prefixes = {'stream', 'obs', 'flux'}

    @classmethod
    def patterns(cls, __id: int) -> typing.List[str]:
        """Generate stream-observer filename patterns for the given ID."""
        if not isinstance(__id, int):
            raise TypeError(
                f"Cannot create stream observer from ID type {type(__id)}"
            ) from None
        return [f'{prefix}{__id:06}.*' for prefix in cls.prefixes]

    def __init__(
        self,
        dataview: datafile.View,
        observables: observable.Quantities,
    ) -> None:
        super().__init__(dataview, observables)
        self._shells = None

    @property
    def shells(self) -> physical.Points:
        """This observer's shell numbers."""
        if self._shells is None:
            self._shells = self._get_axis('shell')
        return self._shells


@etc.autostr
class Point(Observer):
    """An EPREM point observer."""

    prefixes = {'point', 'p_obs'}

    @classmethod
    def patterns(cls, __id: typing.Union[str, int]) -> typing.List[str]:
        """Generate point-observer filename patterns for the given ID."""
        patterns = []
        if not isinstance(__id, (int, str)):
            raise TypeError(
                f"Cannot create point observer from ID type {type(__id)}"
            ) from None
        with contextlib.suppress(ValueError):
            patterns = [f'{prefix}{int(__id):03}.*' for prefix in cls.prefixes]
        if isinstance(__id, str):
            patterns.append(f'{__id}.*')
        return patterns

    def __init__(
        self,
        dataview: datafile.View,
        observables: observable.Quantities,
    ) -> None:
        super().__init__(dataview, observables)
        self._r = None
        self._theta = None
        self._phi = None

    @property
    def r(self):
        """The radial coordinate of this observer."""
        if self._r is None:
            data = float(self['r'][0, 0])
            unit = self['r'].unit
            self._r = physical.scalar(data, unit=unit)
        return self._r

    @property
    def theta(self):
        """The polar coordinate of this observer."""
        if self._theta is None:
            data = float(self['theta'][0, 0])
            unit = self['theta'].unit
            self._theta = physical.scalar(data, unit=unit)
        return self._theta

    @property
    def phi(self):
        """The azimuthal coordinate of this observer."""
        if self._phi is None:
            data = float(self['phi'][0, 0])
            unit = self['phi'].unit
            self._phi = physical.scalar(data, unit=unit)
        return self._phi


@typing.overload
def observer(
    __id: int,
    config: paths.PathLike=None,
    source: paths.PathLike=None,
    system: str=None,
) -> Stream: ...


@typing.overload
def observer(
    __id: str,
    config: paths.PathLike=None,
    source: paths.PathLike=None,
    system: str=None,
) -> Point: ...


def observer(
    __id: typing.Union[str, int],
    config: paths.PathLike=None,
    source: paths.PathLike=None,
    system: str=None,
) -> typing.Union[Stream, Point]:
    """Create an interface to an arbitrary observer file."""
    if not isinstance(__id, (int, str)):
        raise TypeError(
            f"Cannot create observer from ID type {type(__id)}"
        ) from None
    if isinstance(__id, int):
        patterns = (f'{prefix}{__id:06}.*' for prefix in {'obs', 'flux'})
        return _create_observer(Stream, patterns, config, source, system)
    patterns = []
    with contextlib.suppress(ValueError):
        patterns.append(f'p_obs{int(__id):03}.*')
    if isinstance(__id, str):
        patterns.append(f'{__id}.*')
    return _create_observer(Point, tuple(patterns), config, source, system)


def stream(
    __id: int,
    config: paths.PathLike=None,
    source: paths.PathLike=None,
    system: str=None,
) -> Stream:
    """Create an interface to a stream-observer file."""
    patterns = Stream.patterns(__id)
    return _create_observer(Stream, patterns, config, source, system)


def point(
    __id: typing.Union[str, int],
    config: paths.PathLike=None,
    source: paths.PathLike=None,
    system: str=None,
) -> Point:
    """Create an interface to a point-observer file."""
    patterns = Point.patterns(__id)
    return _create_observer(Point, patterns, config, source, system)


def _create_observer(
    __type: typing.Type[ObserverType],
    patterns: typing.Iterable[str],
    config: paths.PathLike=None,
    source: paths.PathLike=None,
    system: str=None,
) -> ObserverType:
    """Internal factory for observer interfaces."""
    directory = paths.fullpath(source or '.', strict=True)
    if not directory.is_dir():
        raise SourcePathError(
            f"Source path {source} is not a directory"
        ) from None
    datapath = _build_datapath(patterns, directory)
    confpath = _build_confpath(directory, config)
    dataview = datafile.view(datapath)
    observables = observable.quantities(datapath, confpath, system)
    return __type(dataview, observables)


@etc.autostr
class Dataset:
    """An interface to a complete EPREM dataset."""

    def __init__(
        self,
        directory: pathlib.Path,
        parameters: parameter.Interface,
        config: parameter.ConfigFile,
        system: metric.System,
    ) -> None:
        self._directory = directory
        self._parameters = parameters
        self._config = config
        self._system = system
        self._reference = None
        self._times = None
        self._shells = None
        self._species = None
        self._energies = None
        self._mus = None
        self._observers = None
        self._streams = None
        self._points = None

    def __str__(self) -> str:
        """Called for str(self)."""
        return str(self.directory)

    @property
    def times(self) -> physical.Coordinates:
        """This dataset's time coordinates."""
        if self._times is None:
            self._times = self._get_reference_attr('times')
        return self._times

    @property
    def shells(self) -> physical.Points:
        """This dataset's shell indices."""
        if self._shells is None:
            self._shells = self._get_reference_attr('shells')
        return self._shells

    @property
    def species(self) -> physical.Symbols:
        """This dataset's species symbols."""
        if self._species is None:
            self._species = self._get_reference_attr('species')
        return self._species

    @property
    def energies(self) -> physical.Coordinates:
        """This dataset's energy coordinates."""
        if self._energies is None:
            self._energies = self._get_reference_attr('energies')
        return self._energies

    @property
    def mus(self) -> physical.Coordinates:
        """This dataset's mu coordinates."""
        if self._mus is None:
            self._mus = self._get_reference_attr('mus')
        return self._mus

    def _get_reference_attr(self, name: str, /):
        """Get an attribute from the reference stream observer."""
        if self._reference is None:
            streams = list(self.streams.values())
            self._reference = streams[0]
        return getattr(self._reference, name)

    @property
    def observers(self) -> dict[str | int, Point | Stream]:
        """A mapping of available observer files.

        Notes
        -----
        This mapping includes all point-observer files, followed by all
        stream-observer files. The reason for this ordering is based on the
        assumtion that users who iterate over all observers (e.g., in a
        visualization script) will want to analyze point observers first, since
        they are most readily comparable to satellite observations. Furthermore,
        most EPREM simulation runs will include far fewer point observers than
        stream observers, which means that this ordering should not be
        inconvenient to users who are more interesting in analyzing stream
        observers.

        See Also
        --------
        `~points`: all available point-observer files
        `~streams`: all available stream-observer files
        """
        if self._observers is None:
            self._observers = {**self.points, **self.streams}
        return self._observers

    @property
    def points(self):
        """A mapping of available point-observer interfaces."""
        if self._points is None:
            obspaths = [
                path
                for prefix in Point.prefixes
                for path in self.directory.glob(f"{prefix}*")
                if path.suffix in datafile.VIEWERS
            ]
            self._points = {
                _get_observer_id(path): self._new_observer(path, Point)
                for path in obspaths
            }
        return self._points

    @property
    def streams(self):
        """A mapping of available stream-observer interfaces."""
        if self._streams is None:
            obspaths = [
                path
                for prefix in Stream.prefixes
                for path in self.directory.glob(f"{prefix}*")
                if path.suffix in datafile.VIEWERS
            ]
            self._streams = {
                _get_observer_id(path): self._new_observer(path, Stream)
                for path in obspaths
            }
        return self._streams

    def _new_observer(
        self,
        path: pathlib.Path,
        obstype: typing.Type[ObserverType]=Observer,
    ) -> ObserverType:
        """Create a new general observer interface."""
        dataview = datafile.view(source=path)
        observables = observable.quantities(
            source=path,
            config=self.config.source,
            system=self.system,
        )
        return obstype(dataview, observables)

    @property
    def directory(self):
        """The full path to EPREM output files."""
        return self._directory

    @property
    def parameters(self):
        """An interface to simulation parameter values."""
        return self._parameters

    @property
    def config(self):
        """An interface to this dataset's raw configuration file."""
        return self._config

    @property
    def system(self):
        """This dataset's metric system."""
        return self._system


def _get_observer_id(path: pathlib.Path):
    """Compute the appropriate observer ID for the given path."""
    stem = path.stem
    prefixes = Stream.prefixes | Point.prefixes
    for prefix in prefixes:
        if stem.startswith(prefix):
            key = stem.lstrip(prefix)
            if prefix in Stream.prefixes:
                return int(key)
            return key
    raise ValueError(path)


def dataset(
    source: paths.PathLike=None,
    config: paths.PathLike=None,
    system: str=None,
) -> Dataset:
    """Create an EPREM dataset interface."""
    directory = paths.fullpath(source or '.', strict=True)
    confpath = _build_confpath(directory, config=config)
    return Dataset(
        directory,
        config=parameter.configfile(confpath),
        parameters=parameter.interface(confpath),
        system=metric.system(system or 'mks'),
    )


class SourcePathError(Exception):
    """Error while creating path to observer data."""


def _build_datapath(
    patterns: typing.Iterable[str],
    directory: pathlib.Path,
) -> pathlib.Path:
    """Create the full path to an observer's data file, if possible."""
    for pattern in patterns:
        filenames = list(directory.glob(pattern))
        if len(filenames) == 1:
            return paths.fullpath(directory / filenames[0], strict=True)
        if len(filenames) > 1:
            raise SourcePathError(
                f"Cannot determine unique data path within {directory}"
                f" from {pattern!r}"
            ) from None
    raise SourcePathError(
        f"Cannot locate observer data within {directory}"
    ) from None


_CONFIG_NAMES = (
    'eprem_input_file',
    '*.cfg',
    '*.ini',
    '*.in',
)

def _build_confpath(
    directory: pathlib.Path,
    config: paths.PathLike=None,
) -> pathlib.Path:
    """Create the full path to the requested config file, if possible."""
    if config is None: # need to guess
        for name in _CONFIG_NAMES:
            if found := next(directory.glob(name), None):
                return found
        raise ValueError(
            f"Cannot guess name of config file in {directory}"
        ) from None
    this = pathlib.Path(config)
    full = (
        directory / config if this.name == config # file name only
        else paths.fullpath(this, strict=True) # full or relative path
    )
    if full.exists():
        return full
    raise ValueError(
        f"Cannot determine path to config file from {config!r}"
    ) from None




import typing

from . import _protocols


MeasurableType = typing.TypeVar('MeasurableType', bound=_protocols.Measurable)
"""Type variable representing explicitly measurable objects."""


MetricType = typing.TypeVar('MetricType', bound=_protocols.HasUnit)
"""Type variable representing objects with a unit."""


MeasuredType = typing.TypeVar('MeasuredType', bound=_protocols.Measured)
"""Type variable representing objects with data and a unit."""


MeasurementType = typing.TypeVar(
    'MeasurementType',
    bound=_protocols.Measurement,
)
"""Type variable representing measurement-like objects."""


ComplexType = typing.TypeVar('ComplexType', bound=_protocols.Complex)
"""Type variable for objects that implement the complex-valued protocol."""


RealType = typing.TypeVar('RealType', bound=_protocols.Real)
"""Type variable for objects that implement the real-valued protocol."""


SequenceType = typing.TypeVar('SequenceType', bound=_protocols.Sequence)
"""Type variable representing real-vauled sequence-like objects."""


ArrayType = typing.TypeVar('ArrayType', bound=_protocols.Array)
"""Type variable representing real-vauled array-like objects."""


ScalarType = typing.TypeVar('ScalarType', bound=_protocols.Scalar)
"""Type variable representing real-vauled scalar objects."""


